%{
(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Readable_types
open Readable_annotations
open Asmprogram
%}

%token LP
%token RP
%token LSB
%token RSB
%token LCB
%token RCB
%token PIPE
%token COLON
%token COMMA
%token SEMICOLON
%token EQ
%token NEQ
%token LT
%token ULT
%token GT
%token UGT
%token LE
%token ULE
%token GE
%token UGE
%token PLUS
%token MINUS
%token STAR
%token DOLLAR
%token <string> REG
%token <string> IDENT
%token <nativeint> UINT
%token ARROW
%token CONS
%token EOF

%token LABEL
%token TYPE
%token BLOCK

%token TAG_BYTE
%token TAG_DWORD

%token INST_MOVL
%token INST_MOVB
%token INST_PUSHL
%token INST_POPL
%token INST_CALL
%token INST_JMP
%token INST_JB
%token INST_JBE
%token INST_JA
%token INST_JAE
%token INST_JZ
%token INST_JNZ
%token INST_RET
%token INST_ADDL
%token INST_SUBL
%token INST_CMPL
%token INST_CMPB
%token INST_MULL
%token INST_IMULL

%token AINST_LONG
%token AINST_TEXT
%token AINST_DATA
%token AINST_GLOBL

%token ROLL
%token UNROLL
%token PACK
%token UNPACK
%token SPLIT
%token CONCAT

%token CAST_TO_TUPLE
%token CAST_TO_ARRAY
%token CAST_TO_STACK

%left PLUS MINUS
%left STAR
%left prec_add
%left prec_mul

%type <Asmprogram.blocks> blocks
%start blocks

%%

tvar_defs :
	| tvar_def tvar_defs
	{ $1 :: $2 }

	| tvar_def
	{ [$1] }

tvar_def :
	| IDENT COLON kind_def SEMICOLON
	{ TV $1, $3 }

kind_def :
	| IDENT
	{
		if "int" = $1 then KInt else
		if "store" = $1 then KStore else
		if "stack" = $1 then KStack else failwith ("Unknown kind : " ^ $1)
	}

store_def:
	| store_elem_def store_def
	{ $1 :: $2 }

	| store_elem_def
	{ [$1] }

	| LP RP
	{ [] }

store_elem_def :
	| LCB int_def ARROW heaptyp_def RCB
	{ SEAtom ($2, $4) }

	| LCB IDENT RCB
	{ SEVar (TV $2) }

c_str_def:
	| c_str_elem_def c_str_def
	{ $1 :: $2 }

	| c_str_elem_def
	{ [$1] }

c_str_elem_def:
	| LCB int_def ARROW code_type_def RCB
	{ ($2, $4) }

typ_def :
	| IDENT LP constructors RP
	{ Name ($1, $3) }

	| IDENT LP RP
	{ Name ($1, []) }

	| ctyp_def
	{ Concrete $1 }

arrtyp_def :
	| typ_def LP int_def RP
	{
		{ at_type = $1; at_len = $3; }
	}

	| typ_def
	{
		{ at_type = $1; at_len = IConst "0x1"; }
	}

heaptyp_def :
	| arrtyp_def
	{ Array $1 }

	| rstacktyp_def
	{ Stack $1 }

code_type_def :
	| IDENT LP constructors RP
	{ NameCode ($1, $3) }

	| IDENT LP RP
	{ NameCode ($1, []) }

	| state_type_def
	{ State $1 }

ctyp_def:
	| tupletyp_def
	{ Tuple $1 }

	| packtyp_def
	{ Pack $1 }

packtyp_def:
	| prefix_def tupletyp_def
	{ 
		{ pt_prefix = $1; pt_tuple = $2; }
	}

	| TAG_BYTE
	{ {
		pt_prefix =
		{
			p_tvs = [TV("_V"), KInt];
			p_cstrts = [];
		};
		pt_tuple = [Instructions.OpSize8, IVar (TV "_V")];
	} }

	| TAG_DWORD
	{ {
		pt_prefix =
		{
			p_tvs = [TV("_V"), KInt];
			p_cstrts = [];
		};
		pt_tuple = [Instructions.OpSize32, IVar (TV "_V")];
	} }

prefix_def:
	| tvar_defs_wrap cstrts_def_wrap
	{
		{ p_tvs = $1; p_cstrts = $2; }
	}

tvar_defs_wrap:
	| LSB tvar_defs
	{ $2 }

	| LSB
	{ [] }

cstrts_def_wrap:
	| RSB cstrts_def
	{ $2 }

	| RSB
	{ [] }

cstrts_def:
	| cstrt_def cstrts_def
	{ $1 :: $2 }

	| cstrt_def
	{ [$1] }

cstrt_def:
	| inte_cstrts_def_wrap RSB
	{
		{ c_ics = $1; c_store = []; c_code_store = []; }
	}

	| inte_cstrts_def_wrap rest1_def_wrap RSB
	{
		let (str, c_str) = $2 in
		{ c_ics = $1; c_store = str; c_code_store = c_str; }
	}

inte_cstrts_def_wrap:
	| LSB inte_cstrts_def PIPE
	{ $2 }

	| LSB
	{ [] }

rest1_def_wrap:
	| store_def PIPE c_str_def
	{ ($1, $3) }

	| store_def
	{ ($1, []) }

	| PIPE c_str_def
	{ ([], $2) }

constructors :
	| constructor COMMA constructors
	{ $1 :: $3 }

	| constructor
	{ [$1] }

constructor :
	| stacktyp_def
	{ CStack $1 }

	| store_def
	{ CStore $1 }

	| int_def
	{ CInt $1 }

opsize_def :
	| UINT
	{
		let i = Nativeint.to_int $1 in
		match i with
		| 8 -> Instructions.OpSize8
		| 16 -> Instructions.OpSize16
		| 32 -> Instructions.OpSize32
		| 48 -> Instructions.OpSize48
		| 64 -> Instructions.OpSize64
		| 128 -> Instructions.OpSize128
		| _ -> failwith "Unknown tuple element size."
	}

tupleelm_def :
	| int_def
	{ Instructions.OpSize32, $1 }

	| int_def COLON opsize_def
	{ $3, $1 }

tupletyp_aux_def :
	| tupleelm_def COMMA tupletyp_aux_def
	{ $1 :: $3 }

	| tupleelm_def
	{ [$1] }

tupletyp_def :
	| LP tupletyp_aux_def RP
	{ $2 }

int_def :
	| imdop
	{ IConst $1 }

	| IDENT
	{ IVar (TV $1) }

inte_def :
	| int_def
	{ IEInte $1 }

	| inte_def STAR inte_def %prec prec_mul
	{ IEArith (A_MUL, $1, $3) }

	| inte_def PLUS inte_def %prec prec_add
	{ IEArith (A_ADD, $1, $3) }

intc_def :
	| inte_def EQ inte_def
	{ ICCmp ((C_EQ, true), $1, $3) }

	| inte_def NEQ inte_def
	{ ICCmp ((C_NEQ, true), $1, $3) }

	| inte_def LT inte_def
	{ ICCmp ((C_LT, true), $1, $3) }

	| inte_def ULT inte_def
	{ ICCmp ((C_LT, false), $1, $3) }

	| inte_def GT inte_def
	{ ICCmp ((C_GT, true), $1, $3) }

	| inte_def UGT inte_def
	{ ICCmp ((C_GT, false), $1, $3) }

	| inte_def LE inte_def
	{ ICCmp ((C_LE, true), $1, $3) }

	| inte_def ULE inte_def
	{ ICCmp ((C_LE, false), $1, $3) }

	| inte_def GE inte_def
	{ ICCmp ((C_GE, true), $1, $3) }

	| inte_def UGE inte_def
	{ ICCmp ((C_GE, false), $1, $3) }


inte_cstrts_def :
	| intc_def
	{ [$1] }

	| intc_def COMMA inte_cstrts_def
	{ $1 :: $3 }

stacktyp_def :
	| int_def CONS stacktyp_def
	{ StkCons ($1, $3) }

	| LSB IDENT RSB
	{ StkVar (TV $2) }

	| LSB RSB
	{ StkNull }

rstacktyp_def :
	| LSB stacktyp_def RSB LP int_def RP
	{
		{ rstk_stk = $2; rstk_freelen = $5; }
	}

regtyp_def :
	| regtyp_elem_def regtyp_def
	{ 
		try
			let (reg, wt) = $1 in
			let _ = List.assoc reg $2 in
			failwith (reg ^ " defined twice.")
		with Not_found ->
			$1 :: $2
	}

	| regtyp_elem_def
	{ [$1] }

regtyp_elem_def :
	| REG COLON int_def SEMICOLON
	{ ($1, $3) }

state_type_def :
	| prefix_def LCB regtyp_def flags_def_wrap
	{
		let rec find names lst =
			match names with
			| [] ->
				failwith "There is a non-defined register."
			| (name, reg_kind) :: names ->
			try
				reg_kind, (List.assoc name lst)
			with Not_found ->
				find names lst
		in

		let find_eax = find [("eax", RK_E); ("ax", RK_X); ("ah", RK_H); ("al", RK_L); ] in
		let find_ecx = find [("ecx", RK_E); ("cx", RK_X); ("ch", RK_H); ("cl", RK_L); ] in
		let find_edx = find [("edx", RK_E); ("dx", RK_X); ("dh", RK_H); ("dl", RK_L); ] in
		let find_ebx = find [("ebx", RK_E); ("bx", RK_X); ("bh", RK_H); ("bl", RK_L); ] in
		let find_esi = find [("esi", RK_E); ("si", RK_X); ] in
		let find_edi = find [("edi", RK_E); ("di", RK_X); ] in
		let find_ebp = find [("ebp", RK_E); ("bp", RK_X); ] in
		let find_esp = find [("esp", RK_E); ("sp", RK_X); ] in

		let create_regs regs =
		{
			cs = CodeSeg SegSize32;
			eax = find_eax regs;
			ecx = find_ecx regs;
			edx = find_edx regs;
			ebx = find_ebx regs;
			esi = find_esi regs;
			edi = find_edi regs;
			ebp = find_ebp regs;
			esp = find_esp regs;
			flags_opt = $4;
		}
		
		in

		let prefix = $1 in
		let reg_typ = create_regs $3 in
		{ st_prefix = prefix; st_registers = reg_typ; }
	}

flags_def_wrap:
	| RCB flags_def
	{ Some $2 }

	| RCB
	{ None }

flags_def:
	| LCB regtyp_def op_size_wrap
	{
		let rec find name lst =
		try
			List.assoc name lst
		with Not_found ->
			failwith "There is a non-defined flag component."
		in

		{
			f_rst = find "f_rst" $2;
			f_op1 = find "f_op1" $2;
			f_op2 = find "f_op2" $2;
			f_size = $3;
		}
	}

op_size_wrap:
	| RCB LP UINT RP
	{
		match $3 with
		| 8n -> Instructions.OpSize8
		| 16n -> Instructions.OpSize16
		| 32n -> Instructions.OpSize32
		| 48n -> Instructions.OpSize48
		| 64n -> Instructions.OpSize64
		| 128n -> Instructions.OpSize128
		| _ ->
			failwith ("Unknown operand size: " ^ (Nativeint.to_string $3))
	}

	| RCB
	{ Instructions.OpSize32 }

insts:
	| inst insts
	{ $1 :: $2 }

	| inst
	{ [$1] }

inst:
	| raw_inst SEMICOLON
	{ IRaw ($1, []) }

	| raw_inst ARROW ids SEMICOLON
	{ IRaw ($1, $3) }

	| asm_inst SEMICOLON
	{ $1 }

	| pseudo_inst SEMICOLON
	{ IAnnot $1 }

	| IDENT COLON
	{ ILabel $1 }


raw_inst:
	| INST_MOVL operand COMMA operand
	{ Inst2 ("movl", $2, $4) }

	| INST_MOVB operand COMMA operand
	{ Inst2 ("movb", $2, $4) }

	| INST_PUSHL operand
	{ Inst1 ("pushl", $2) }

	| INST_POPL operand
	{ Inst1 ("popl", $2) }

	| INST_RET
	{ Inst0 "ret" }

	| INST_CALL operand
	{ Inst1 ("call", to_signed $2) }

	| INST_JMP operand
	{ Inst1 ("jmp", to_signed $2) }

	| INST_JB operand
	{ Inst1 ("jb", to_signed $2) }

	| INST_JBE operand
	{ Inst1 ("jbe", to_signed $2) }

	| INST_JA operand
	{ Inst1 ("ja", to_signed $2) }

	| INST_JAE operand
	{ Inst1 ("jae", to_signed $2) }

	| INST_JZ operand
	{ Inst1 ("jz", to_signed $2) }

	| INST_JNZ operand
	{ Inst1 ("jnz", to_signed $2) }

	| INST_ADDL operand COMMA operand
	{ Inst2 ("addl", $2, $4) }

	| INST_SUBL operand COMMA operand
	{ Inst2 ("subl", $2, $4) }

	| INST_CMPL operand COMMA operand
	{ Inst2 ("cmpl", $2, $4) }

	| INST_CMPB operand COMMA operand
	{ Inst2 ("cmpb", $2, $4) }

	| INST_MULL operand
	{ Inst1 ("mull", $2) }

	| INST_IMULL operand COMMA operand COMMA operand
	{ Inst3 ("imull", $2, $4, $6) }

asm_inst:
	| AINST_TEXT
	{ IRaw (Inst0 (".text"), []) }

	| AINST_DATA
	{ IRaw (Inst0 (".data"), []) }

	| AINST_GLOBL IDENT
	{ IRaw (Inst0 (".globl " ^ $2), []) }

	| AINST_LONG imdop
	{ IRaw (Inst0 (".long " ^ $2), []) }

ids:
	| IDENT COMMA ids
	{ ($1) :: $3 }

	| IDENT
	{ [$1] }

pseudo_inst:
	| ROLL int_def COMMA IDENT LP constructors RP
	{
		let nt = ($4, $6) in
		A_ROLL ($2, nt)
	}

	| ROLL int_def COMMA IDENT LP RP
	{
		let nt = ($4, []) in
		A_ROLL ($2, nt)
	}

	| UNROLL int_def
	{ A_UNROLL $2 }

	| PACK int_def COMMA packtyp_def LSB constructors RSB
	{ A_PACK ($2, $4, $6) }

	| PACK int_def COMMA packtyp_def
	{ A_PACK ($2, $4, []) }

	| UNPACK int_def ARROW ids
	{ A_UNPACK ($4, $2) }

	| UNPACK int_def
	{ A_UNPACK ([], $2) }

	| SPLIT int_def COMMA int_def ARROW ids
	{ A_SPLIT ($6, $2, $4) }

	| SPLIT int_def COMMA int_def
	{ A_SPLIT ([], $2, $4) }

	| CONCAT int_def COMMA int_def ARROW ids
	{ A_CONCAT ($6, $2, $4) }

	| CONCAT int_def COMMA int_def
	{ A_CONCAT ([], $2, $4) }

	| CAST_TO_TUPLE int_def ARROW ids
	{ A_CAST_TO_TUPLE ($4, $2) }

	| CAST_TO_TUPLE int_def
	{ A_CAST_TO_TUPLE ([], $2) }

	| CAST_TO_ARRAY int_def ARROW ids
	{ A_CAST_TO_ARRAY ($4, $2) }

	| CAST_TO_ARRAY int_def
	{ A_CAST_TO_ARRAY ([], $2) }

	| CAST_TO_STACK int_def ARROW ids
	{ A_CAST_TO_STACK ($4, $2) }

	| CAST_TO_STACK int_def
	{ A_CAST_TO_STACK ([], $2) }

operand:
	| imdop
	{
		let sign = Util.get_sign $1 in
		OpImd ($1, sign)
	}

	| REG
	{ OpReg $1 }

	| memop EQ LSB memannot RSB
	{ OpMem ($1, $4) }

	| LSB memannot RSB EQ memop
	{ OpMem ($5, $2) }

imdop:
	| symbol_or_int_arith
	{ $1 }

int:
	| MINUS UINT
	{ Nativeint.neg $2 }

	| UINT
	{ $1 }

memop:
	| symbol_or_int
	{ $1 }

	| LP REG RP
	{ "(%" ^ $2 ^ ")" }

	| LP REG COMMA REG RP
	{ "(%" ^ $2 ^ ", %" ^ $4 ^ ")" }

	| int LP REG RP
	{
		let s = Nativeint.to_string $1 in
		s ^ "(%" ^ $3 ^ ")"
	}

	| int LP REG COMMA REG RP
	{
		let s = Nativeint.to_string $1 in
		s ^ "(%" ^ $3 ^ ", %" ^ $5 ^ ")"
	}

memannot:
	| int_def
	{ ($1, 0n) }

	| int_def PLUS int
	{ ($1, $3) }
/*
	| int_def
	{ ($1, 0n, IConst "0", 0n) }

	| int_def PLUS int
	{ ($1, 0n, IConst "0", $3) }

	| int_def PLUS int STAR int_def PLUS int
	{ ($1, $3, $5, $7) }
*/

symbol_or_int_arith:
	| DOLLAR symbol_or_int
	{ $2 }

	| LP symbol_or_int_arith RP
	{ $2 }

	| symbol_or_int_arith PLUS symbol_or_int_arith
	{ $1 ^ "+" ^ $3 }

	| symbol_or_int_arith MINUS symbol_or_int_arith
	{ $1 ^ "-" ^ $3 }

	| symbol_or_int_arith STAR symbol_or_int_arith
	{ $1 ^ "*" ^ $3 }

symbol_or_int:
	| IDENT
	{ $1 }

	| int
	{ Nativeint.to_string $1 }

blocks:
	| block blocks
	{ $1 :: $2 }

	| EOF
	{ [] }

block:
	| LABEL IDENT EQ heaptyp_def
	{
		ALabelType { l_name = $2; l_type = LTData $4; }
	}

	| LABEL IDENT EQ state_type_def
	{
		ALabelType { l_name = $2; l_type = LTCode $4; }
	}

	| TYPE IDENT LSB tvar_defs RSB EQ ctyp_def
	{
		let ntd = { ntd_tvs = $4; ntd_body = NtdConcrete $7; } in
		ANamedType { nt_name = $2; nt_definition = ntd; }
	}

	| TYPE IDENT LSB tvar_defs RSB EQ state_type_def
	{
		let ntd = { ntd_tvs = $4; ntd_body = NtdState $7; } in
		ANamedType { nt_name = $2; nt_definition = ntd; }
	}

	| TYPE IDENT EQ ctyp_def
	{
		let ntd = { ntd_tvs = []; ntd_body = NtdConcrete $4; } in
		ANamedType { nt_name = $2; nt_definition = ntd; }
	}

	| TYPE IDENT EQ state_type_def
	{
		let ntd = { ntd_tvs = []; ntd_body = NtdState $4; } in
		ANamedType { nt_name = $2; nt_definition = ntd; }
	}

	| insts
	{
		AProgram ("", $1)
	}

	| BLOCK IDENT LCB insts RCB
	{
		AProgram ($2, $4)
	}
