(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Machine
open Instructions
open Instructions_util

open Disassemble_prefix
open Disassemble_tables

open Util

(***********************************************
	Disassemble Common Data and Routines
 ***********************************************)

type disassemble_context =
{
	dctx_op_attr	: operand_size_attribute;
	dctx_ad_attr	: address_size_attribute;
	dctx_prefixes	: prefixes;
}

let get_mem_op op_size base_opt disp_opt scaled_index_opt =
	Op_Mem {
		mop_disp_opt = disp_opt;
		mop_base_opt = base_opt;
		mop_scaled_index_opt = scaled_index_opt;
		mop_size = op_size;
	}

let op_attr_override sz =
	match sz with
	| OpSizeAttr16 -> OpSizeAttr32
	| OpSizeAttr32 -> OpSizeAttr16

let ad_attr_override sz =
	match sz with
	| AdSizeAttr16 -> AdSizeAttr32
	| AdSizeAttr32 -> AdSizeAttr16

let has_op_override prefixes = prefixes.g3_opt <> None
let has_ad_override prefixes = prefixes.g4_opt <> None

let get_cc index =
	match index with
        | 0x0 -> CC_O
        | 0x1 -> CC_NO
        | 0x2 -> CC_ULT
        | 0x3 -> CC_UGE
        | 0x4 -> CC_Z
        | 0x5 -> CC_NZ
        | 0x6 -> CC_ULE
        | 0x7 -> CC_UGT
        | 0x8 -> CC_S
        | 0x9 -> CC_NS
        | 0xA -> CC_PE
        | 0xB -> CC_PO
        | 0xC -> CC_SLT
        | 0xD -> CC_SGE
        | 0xE -> CC_SLE
        | 0xF -> CC_SGT
	| _ -> failwithf "Invalid condition code index value = %d." index

let get_genreg op_size reg =
	match (op_size, reg) with
	| ( OpSize8, 0) -> (Reg8 AL)
	| (OpSize16, 0) -> (Reg16 AX)
	| (OpSize32, 0) -> (Reg32 EAX)
	| ( OpSize8, 1) -> (Reg8 CL)
	| (OpSize16, 1) -> (Reg16 CX)
	| (OpSize32, 1) -> (Reg32 ECX)
	| ( OpSize8, 2) -> (Reg8 DL)
	| (OpSize16, 2) -> (Reg16 DX)
	| (OpSize32, 2) -> (Reg32 EDX)
	| ( OpSize8, 3) -> (Reg8 BL)
	| (OpSize16, 3) -> (Reg16 BX)
	| (OpSize32, 3) -> (Reg32 EBX)
	| ( OpSize8, 4) -> (Reg8 AH)
	| (OpSize16, 4) -> (Reg16 SP)
	| (OpSize32, 4) -> (Reg32 ESP)
	| ( OpSize8, 5) -> (Reg8 CH)
	| (OpSize16, 5) -> (Reg16 BP)
	| (OpSize32, 5) -> (Reg32 EBP)
	| ( OpSize8, 6) -> (Reg8 DH)
	| (OpSize16, 6) -> (Reg16 SI)
	| (OpSize32, 6) -> (Reg32 ESI)
	| ( OpSize8, 7) -> (Reg8 BH)
	| (OpSize16, 7) -> (Reg16 DI)
	| (OpSize32, 7) -> (Reg32 EDI)
	| _ -> failwithf "Invalid general register operand value = %d." reg

let get_genreg_op op_size reg =
	Op_Reg (get_genreg op_size reg)

let get_segreg reg =
	match reg with
	| 0 -> (Regseg ES)
	| 1 -> (Regseg CS)
	| 2 -> (Regseg SS)
	| 3 -> (Regseg DS)
	| 4 -> (Regseg FS)
	| 5 -> (Regseg GS)
	| 6 | 7 -> failwithf "Reserved segment register operand value = %d." reg
	| _ -> failwithf "Invalid segment register operand value = %d." reg

let get_segreg_op reg =
	Op_Reg (get_segreg reg)

let get_operand_I dctx ot seg off =
	let op_size = ot_to_op_size ot dctx.dctx_op_attr in
	let off, imd = match op_size with
	| OpSize8  -> off + ( 8 / bits_per_byte), get_sint8 seg off
	| OpSize16 -> off + (16 / bits_per_byte), get_sint16 seg off
	| OpSize32 -> off + (32 / bits_per_byte), get_sint32 seg off
	| _ -> failwithf "Non supported immediate value size = %a." output_opsize op_size in
	off, Op_Imd (imd, op_size)

let get_operand_J dctx ot seg off =
	(* same as get_operand_I. The difference is that J is relative addressing. *)
	get_operand_I dctx ot seg off

let get_operand_O dctx ot seg off =
	let op_size = ot_to_op_size ot dctx.dctx_op_attr in
	let off, imd = match op_size with
	| OpSize8  -> off + ( 8 / bits_per_byte), get_sint8 seg off
	| OpSize16 -> off + (16 / bits_per_byte), get_sint16 seg off
	| OpSize32 -> off + (32 / bits_per_byte), get_sint32 seg off
	| _ -> failwithf "Non supported immediate value size = %a." output_opsize op_size in
	off, get_mem_op op_size None (Some imd) None

let reserved_inst f =
	Format.kfprintf
	(fun _ ->
		let str = Format.flush_str_formatter () in
		(fun dctx seg off -> failwith str)
	) Format.str_formatter f
