(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Printf
open Elf_header
open Elf_section
open Elf_segment

let output_e_ident_class o ei_class =
	let str = match ei_class with
		| ELFCLASS32 -> "ELFCLASS32"
		| ELFCLASS64 -> "ELFCLASS64"
	in
	fprintf o "%s" str

let output_e_ident_data o ei_data =
	let str = match ei_data with
		| ELFDATA2LSB -> "ELFDATA2LSB"
		| ELFDATA2MSB -> "ELFDATA2MSB"
	in
	fprintf o "%s" str

let output_e_ident_version o ei_version =
	let str = match ei_version with
		| EV_CURRENT -> "EV_CURRENT"
	in
	fprintf o "%s" str

let output_e_ident o e_ident =
	fprintf o " ei_class = %a\n" output_e_ident_class e_ident.ei_class;
	fprintf o " ei_data = %a\n" output_e_ident_data e_ident.ei_data;
	fprintf o " ei_version = %a\n" output_e_ident_version e_ident.ei_version

let output_e_type o e_type =
	let str = match e_type with
		| ET_REL -> "ET_REL"
		| ET_EXEC -> "ET_EXEC"
		| ET_DYN -> "ET_DYN"
		| ET_CORE -> "ET_CORE"
	in
	fprintf o "%s" str

let output_e_machine o e_machine =
	let str = match e_machine with
		| EM_M32 -> "EM_M32"
		| EM_SPARC -> "EM_SPARC"
		| EM_386 -> "EM_386"
		| EM_68K -> "EM_68K"
		| EM_88K -> "EM_88K"
		| EM_860 -> "EM_860"
		| EM_MIPS -> "EM_MIPS"
		| EM_MIPS_RS4_BE -> "EM_MIPS_RS4_BE"
	in
	fprintf o "%s" str

let output_e_version o e_version =
	let str = match e_version with
		| EV_CURRENT -> "EV_CURRENT"
	in
	fprintf o "%s" str

let output_elf32_half o v = fprintf o "0x%04nx (%nd)" v v
let output_elf32_word o v = fprintf o "0x%08nx (%nd)" v v
let output_elf32_addr o v = fprintf o "0x%08nx (%nd)" v v
let output_elf32_off o v = fprintf o "0x%08nx (%nd)" v v

let output_header o header =
	fprintf o "e_ident = \n%a" output_e_ident header.e_ident;
	fprintf o "e_type = %a\n" output_e_type header.e_type;
	fprintf o "e_machine = %a\n" output_e_machine header.e_machine;
	fprintf o "e_version = %a\n" output_e_version header.e_version;
	fprintf o "e_entry = %a\n" output_elf32_addr header.e_entry;
	fprintf o "e_phoff = %a\n" output_elf32_off header.e_phoff;
	fprintf o "e_shoff = %a\n" output_elf32_off header.e_shoff;
	fprintf o "e_flags = %a\n" output_elf32_word header.e_flags;
	fprintf o "e_ehsize = %a\n" output_elf32_half header.e_ehsize;
	fprintf o "e_phentsize = %a\n" output_elf32_half header.e_phentsize;
	fprintf o "e_phnum = %a\n" output_elf32_half header.e_phnum;
	fprintf o "e_shentsize = %a\n" output_elf32_half header.e_shentsize;
	fprintf o "e_shnum = %a\n" output_elf32_half header.e_shnum;
	fprintf o "e_shstrndx = %a\n" output_elf32_half header.e_shstrndx

let output_sh_type o sh_type =
	let str = match sh_type with
		| SHT_PROGBITS -> "PROGBITS"
		| SHT_SYMTAB -> "SYMTAB"
		| SHT_STRTAB -> "STRTAB"
		| SHT_RELA -> "RELA"
		| SHT_HASH -> "HASH"
		| SHT_DYNAMIC -> "DYNAMIC"
		| SHT_NOTE -> "NOTE"
		| SHT_NOBITS -> "NOBITS"
		| SHT_REL -> "REL"
		| SHT_SHLIB -> "SHLIB"
		| SHT_DYNSYM -> "DYNSYM"
		| SHT_VERSYM -> "GNU?VERSYM"
		| SHT_VERNEED -> "GNU?VERNEED"
	in
	fprintf o "%s" str

let output_sh_flags o sh_flags =
	let a = if sh_flags.shf_alloc then "A" else "_" in
	let w = if sh_flags.shf_write then "W" else "_" in
	let x = if sh_flags.shf_execinstr then "X" else "_" in
	fprintf o "%s%s%s" a w x

let output_section_header o section_header =
	fprintf o "sh_name = %a\n" output_elf32_word section_header.sh_name;
	fprintf o "sh_type = %a\n" output_sh_type section_header.sh_type;
	fprintf o "sh_flags = %a\n" output_sh_flags section_header.sh_flags;
	fprintf o "sh_addr = %a\n" output_elf32_addr section_header.sh_addr;
	fprintf o "sh_offset = %a\n" output_elf32_off section_header.sh_offset;
	fprintf o "sh_size = %a\n" output_elf32_word section_header.sh_size;
	fprintf o "sh_link = %a\n" output_elf32_word section_header.sh_link;
	fprintf o "sh_info = %a\n" output_elf32_word section_header.sh_info;
	fprintf o "sh_addralign = %a\n" output_elf32_word section_header.sh_addralign;
	fprintf o "sh_entsize = %a\n" output_elf32_word section_header.sh_entsize

let output_p_type o p_type =
	let str = match p_type with
	| PT_LOAD -> "PT_LOAD"
	| PT_DYNAMIC -> "PT_DYNAMIC"
	| PT_INTERP -> "PT_INTERP"
	| PT_NOTE -> "PT_NOTE"
	| PT_SHLIB -> "PT_SHLIB"
	| PT_PHDR -> "PT_PHDR"
	| PT_GNU_EH_FRAME -> "PT_GNU_EH_FRAME"
	| PT_STACK -> "PT_STACK"
	in
	fprintf o "%s" str

let output_p_flags o p_flags =
	let r = if p_flags.phf_r then "R" else "_" in
	let w = if p_flags.phf_w then "W" else "_" in
	let x = if p_flags.phf_x then "X" else "_" in
	fprintf o "%s%s%s" r w x

let output_program_header o program_header =
	fprintf o "p_type = %a\n" output_p_type program_header.p_type;
	fprintf o "p_offset = %a\n" output_elf32_off program_header.p_offset;
	fprintf o "p_vaddr = %a\n" output_elf32_addr program_header.p_vaddr;
	fprintf o "p_paddr = %a\n" output_elf32_addr program_header.p_paddr;
	fprintf o "p_filesz = %a\n" output_elf32_word program_header.p_filesz;
	fprintf o "p_memsz = %a\n" output_elf32_word program_header.p_memsz;
	fprintf o "p_flags = %a\n" output_p_flags program_header.p_flags;
	fprintf o "p_align = %a\n" output_elf32_word program_header.p_align
