(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Readable_types

open Util

open Format

(**********************************************
	Routine for Printing Readable_types
 **********************************************)

let rec output_list ppf lst delim output_elm =
	let rec aux first lst =
		match lst with
		| [] -> ()
		| elm :: lst ->
		(if first then
			fprintf ppf "%a" output_elm elm
		else
			fprintf ppf "%s@ %a" delim output_elm elm);
		aux false lst
	in
	aux true lst

let rec output_kind ppf k =
	match k with
	| KStack -> fprintf ppf "STACK"
	| KInt -> fprintf ppf "INT"
	| KStore -> fprintf ppf "STORE"

and output_type ppf t =
	match t with
	| Name nt ->
		fprintf ppf "%a" output_nt nt
	| Concrete (ct) ->
		fprintf ppf "%a" output_ct ct

and output_at ppf at =
	let t = at.at_type in
	let len = at.at_len in

	fprintf ppf "%a" output_type t;

	if len <> (IConst "1") || len <> (IConst "0x1") then
		fprintf ppf "[%a]" output_inte len

and output_ht ppf = function
	| Array at ->
		output_at ppf at
	| Stack rstk ->
		output_rstk ppf rstk

and output_code ppf ct =
	match ct with
	| NameCode nt ->
		fprintf ppf "%a" output_nt nt
	| State st ->
		fprintf ppf "%a" output_state st

and output_nt ppf nt =
	let (n, cs) = nt in
	fprintf ppf "%s(" n;
	fprintf ppf "%a" output_cs cs;
	fprintf ppf ")"

and output_ct ppf ct =
	match ct with
	(* De Bruijn's notation! *)
	| Pack pt ->
		fprintf ppf "%a" output_pt pt
	| Tuple tt ->
		fprintf ppf "%a" output_tt tt

and output_tt ppf tt =
	fprintf ppf "@[<3><%a>@]" output_tes tt

and output_tes ppf tes =
	output_list ppf tes "," output_te

and output_te ppf (sz, inte) =
        match sz with
        | Instructions.OpSize8 ->
                fprintf ppf "B8(%a)" output_inte inte
        | Instructions.OpSize16 ->
                fprintf ppf "B16(%a)" output_inte inte
        | Instructions.OpSize32 ->
                fprintf ppf "%a" output_inte inte
        | Instructions.OpSize48 ->
                fprintf ppf "B48(%a)" output_inte inte
        | Instructions.OpSize64 ->
                fprintf ppf "B64(%a)" output_inte inte
        | Instructions.OpSize128 ->
                fprintf ppf "B128(%a)" output_inte inte

and output_pt ppf pt =
	let { pt_prefix = prefix; pt_tuple = tt; } = pt in
	fprintf ppf "@[<1>Exists@,%a@ %a@]" output_prefix prefix output_tt tt

and output_state ppf st =
	let { st_prefix = prefix; st_registers = regs_type; } = st in
	fprintf ppf "@[<1>Forall@,%a@ %a@]" output_prefix prefix output_regs_type regs_type

and output_prefix ppf prefix =
	let tvs = prefix.p_tvs in
	let cstrts = prefix.p_cstrts in

	fprintf ppf "@[<1>[";
	List.iter (fun (tv, kind) -> fprintf ppf "%a:%a;@ " output_tv tv output_kind kind) tvs;
	fprintf ppf "]@ ";
	fprintf ppf "%a" output_cstrts cstrts;
	fprintf ppf "@]";

and output_cstrts ppf cstrts =
	List.iter (fun cstrt -> fprintf ppf "%a@ " output_cstrt cstrt) cstrts

and output_cstrt ppf cstrt =
	let ics = cstrt.c_ics in
	let store = cstrt.c_store in
	let code_store = cstrt.c_code_store in

	fprintf ppf "@[<1>[";
	fprintf ppf "%a" output_inte_cstrts ics;
	if ics <> [] && (store <> [] || code_store <> []) then fprintf ppf "|@ ";
	fprintf ppf "%a" output_store store;
	if store <> [] && code_store <> [] then fprintf ppf "|@ ";
	if code_store <> [] then
		fprintf ppf "%a" output_code_store code_store;
	fprintf ppf "]@]";

and output_cs ppf cs =
	List.iter (fprintf ppf "%a " output_cstr) cs

and output_regs_type ppf regs_type =
	fprintf ppf "@[<1>{@,";
(*
	fprintf ppf "eax = %a;@ " output_inte reg_type.eax;
	fprintf ppf "ecx = %a;@ " output_inte reg_type.ecx;
	fprintf ppf "edx = %a;@ " output_inte reg_type.edx;
	fprintf ppf "ebx = %a;@ " output_inte reg_type.ebx;
	fprintf ppf "esi = %a;@ " output_inte reg_type.esi;
	fprintf ppf "edi = %a;@ " output_inte reg_type.edi;
	fprintf ppf "ebp = %a;@ " output_inte reg_type.ebp;
	fprintf ppf "esp = %a;@ " output_inte reg_type.esp;
*)
	fprintf ppf "eax = %a;@ " output_reg_type regs_type.eax;
	fprintf ppf "ecx = %a;@ " output_reg_type regs_type.ecx;
	fprintf ppf "edx = %a;@ " output_reg_type regs_type.edx;
	fprintf ppf "ebx = %a;@ " output_reg_type regs_type.ebx;
	fprintf ppf "esi = %a;@ " output_reg_type regs_type.esi;
	fprintf ppf "edi = %a;@ " output_reg_type regs_type.edi;
	fprintf ppf "ebp = %a;@ " output_reg_type regs_type.ebp;
	fprintf ppf "esp = %a;@ " output_reg_type regs_type.esp;
	fprintf ppf "}@]";

and output_reg_type ppf (reg_kind, inte) =
	let suffix = match reg_kind with
		| RK_L -> "(L)"
		| RK_H -> "(H)"
		| RK_X -> "(X)"
		| RK_E -> ""
	in
	fprintf ppf "%a%s" output_inte inte suffix

(*
and output_wts ppf wts =
	output_list ppf wts "," output_inte
*)

and output_stk ppf = function
	| StkVar (tv) ->
		fprintf ppf "%a" output_tv tv
	| StkNull ->
		fprintf ppf "$"
	| StkCons (inte, st')->
		fprintf ppf "%a@ ::@ %a" output_inte inte output_stk st'

and output_rstk ppf rstk =
	let stk = rstk.rstk_stk in
	let freelen = rstk.rstk_freelen in

	fprintf ppf "%a(%a)" output_stk stk output_inte freelen

and output_tv ppf tv =
	let TV(str) = tv in
	fprintf ppf "%s" str
		
and output_inte ppf = function
	| IConst (s) ->
		fprintf ppf "%s" s
	| IVar (tv) ->
		fprintf ppf "%a" output_tv tv

and output_aop ppf = function
	| A_ADD ->
		fprintf ppf "+"
	| A_MUL ->
		fprintf ppf "*"

and output_cop_raw ppf = function
	| C_EQ ->
		fprintf ppf "="
	| C_NEQ ->
		fprintf ppf "<>"
	| C_LT ->
		fprintf ppf "<"
	| C_GT ->
		fprintf ppf ">"
	| C_LE ->
		fprintf ppf "<="
	| C_GE ->
		fprintf ppf ">="

and output_cop ppf cop =
	let (cop_r, signed) = cop in
	fprintf ppf "%a%s" output_cop_raw cop_r (if signed then "" else "u")

and output_inte_expr ppf = function
	| IEInte inte ->
		fprintf ppf "%a" output_inte inte
	| IEArith (aop, inte_e1, inte_e2) ->
		fprintf ppf "@[(%a@ %a@ %a)@]" output_inte_expr inte_e1 output_aop aop output_inte_expr inte_e2

and output_inte_cstrt ppf = function
	| ICCmp (cop, inte_e1, inte_e2) ->
		fprintf ppf "@[<5>{%a@ %a@ %a}@;@,@]"
			output_inte_expr inte_e1
			output_cop cop
			output_inte_expr inte_e2

and output_inte_cstrts ppf ics =
	List.iter (output_inte_cstrt ppf) ics

and output_store ppf str =
	List.iter (output_store_elem ppf) str

and output_store_elem ppf = function
	| SEAtom atom ->
		fprintf ppf "%a@ " output_atom atom
	| SEVar tv ->
		fprintf ppf "{%a}@ " output_tv tv

and output_atom ppf (inte, ht) =
	fprintf ppf "@[<5>{%a@ -->@ %a}@]" output_inte inte output_ht ht

and output_code_store ppf c_str =
	let output_aux ppf e =
		let (inte, st) = e in
		fprintf ppf "@[<5>{%a@ -->@ %a}@]" output_inte inte output_code st
	in
	List.iter (output_aux ppf) c_str

and output_cstr ppf a =
	match a with
	| CStack st -> fprintf ppf "%a:stack" output_stk st
	| CInt inte -> fprintf ppf "%a:int" output_inte inte
	| CStore str -> fprintf ppf "%a:store" output_store str
