(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Machine
open Read_elf_common

open Readable_types
open Readable_annotations
open Read_types_common
open Util
open Big_int_util

(* Machine-dependent (for IA32) *)

let name_of_section_for_annots = ".annots"

let read_mem_annot seg off =
	let off, base = read_integer seg off in
(*
	let off, scale = read_nativeint seg off in
	let off, index = read_integer seg off in
*)
	let off, offset = read_nativeint seg off in

(*
	let scale = big_int_of_nativeint_unsigned scale in
*)
	let offset = big_int_of_nativeint_signed offset in

	let annot = A_ADDRESS
	{
		ad_base = base;
(*
		ad_scale = scale;
		ad_index = index;
*)
		ad_offset = offset;
	} in
	off, A_ATTR annot

let read_sign_annot seg off =
	off, A_ATTR A_SIGNED

let read_newnames seg off =
	let off, names = read_list read_string seg off in
	let annot = A_NEWNAMES names in
	off, A_ATTR annot

let read_roll seg off =
	let off, inte = read_integer seg off in
	let off, nt = read_named_type seg off in
	let annot = A_ROLL (inte, nt) in
	off, A_INST annot

let read_unroll seg off =
	let off, inte = read_integer seg off in
	let annot = A_UNROLL inte in
	off, A_INST annot

let read_tvs seg off =
	read_list read_tv seg off

let read_pack seg off =
	let off, inte = read_integer seg off in
	let off, pt = read_pack_type seg off in
	let off, cs = read_constructors seg off in

	let annot = A_PACK (inte, pt, cs) in
	off, A_INST annot

let read_unpack seg off =
	let off, inte = read_integer seg off in
	let off, dbg_info = read_list read_string seg off in
	let annot = A_UNPACK (dbg_info, inte) in
	off, A_INST annot

let read_split seg off =
	let off, inte1 = read_integer seg off in
	let off, inte2 = read_integer seg off in
	let off, dbg_info = read_list read_string seg off in
	let annot = A_SPLIT (dbg_info, inte1, inte2) in
	off, A_INST annot

let read_concat seg off =
	let off, inte1 = read_integer seg off in
	let off, inte2 = read_integer seg off in
	let off, dbg_info = read_list read_string seg off in
	let annot = A_CONCAT (dbg_info, inte1, inte2) in
	off, A_INST annot

let read_tuple_split seg off =
	let off, inte1 = read_integer seg off in
	let off, inte2 = read_integer seg off in
	let off, dbg_info = read_list read_string seg off in
	let annot = A_TUPLE_SPLIT (dbg_info, inte1, inte2) in
	off, A_INST annot

let read_tuple_concat seg off =
	let off, inte1 = read_integer seg off in
	let off, inte2 = read_integer seg off in
	let annot = A_TUPLE_CONCAT (inte1, inte2) in
	off, A_INST annot

let read_cast_to_tuple seg off =
	let off, inte = read_integer seg off in
	let off, dbg_info = read_list read_string seg off in
	let annot = A_CAST_TO_TUPLE (dbg_info, inte) in
	off, A_INST annot

let read_cast_to_array seg off =
	let off, inte = read_integer seg off in
	let off, dbg_info = read_list read_string seg off in
	let annot = A_CAST_TO_ARRAY (dbg_info, inte) in
	off, A_INST annot

let read_cast_to_stack seg off =
	let off, inte = read_integer seg off in
	let off, dbg_info = read_list read_string seg off in
	let annot = A_CAST_TO_STACK (dbg_info, inte) in
	off, A_INST annot

let read_annot seg off =
	let off, c = get_char seg off in

	match c with
	| 'M' ->
		read_mem_annot seg off
	| '-' ->
		read_sign_annot seg off
	| 'D' ->
		read_newnames seg off
	| 'R' ->
		read_roll seg off
	| 'U' ->
		read_unroll seg off
	| 'P' ->
		read_pack seg off
	| 'O' ->
		read_unpack seg off
	| 'S' ->
		read_split seg off
	| 'C' ->
		read_concat seg off
	| 'L' ->
		read_tuple_split seg off
	| 'N' ->
		read_tuple_concat seg off
	| 'E' ->
		read_cast_to_tuple seg off
	| 'Y' ->
		read_cast_to_array seg off
	| 'K' ->
		read_cast_to_stack seg off
	| _ ->
		failwithf "Unknown annoation '%c'" c

let add_a adder ad a m =
	let all_annot =
	try
		NativeintMap.find ad m
	with Not_found ->
		{ annot_insts = []; annot_attrs = []; } in

	let all_annot = adder a all_annot in
	NativeintMap.add ad all_annot m

let add_ainst ainst all_annot =
	let insts = ainst :: all_annot.annot_insts in
	{ all_annot with annot_insts = insts; }

let add_aattr aattr all_annot =
	let attrs = aattr :: all_annot.annot_attrs in
	{ all_annot with annot_attrs = attrs; }

let add_annotation ad annot m =
	match annot with
	| A_INST ainst ->
		add_a add_ainst ad ainst m
	| A_ATTR aattr ->
		add_a add_aattr ad aattr m

let rec read_annotations seg off m =
	if off >= seg.seg_size then off, m
	else

	let off, c = get_char seg off in
	match c with
	| 'E' ->
		off, m
	| 'A' ->

	let ad = get_word seg off in

	let off = off + size_of_word in

	let off, annot = read_annot seg off in

	let m = add_annotation ad annot m in

	read_annotations seg off m

	| _ ->
		failwithf "Unknown annotation mark '%c'" c

let rec read_annotations_block seg off m =
	if off >= seg.seg_size then m
	else

	let ad = get_word seg off in
	let off = off + size_of_word in

	let off, m = read_annotations seg off m in

	read_annotations_block seg off m

let reverse_annot annot =
	let attrs = List.rev annot.annot_attrs in
	let insts = List.rev annot.annot_insts in
	{ annot_attrs = attrs; annot_insts = insts; }

let read_annotations_wrap seg =
	let m = read_annotations_block seg 0 NativeintMap.empty in

	NativeintMap.fold (fun ad annot lst ->
		let annot = reverse_annot annot in
		(ad, annot) :: lst
	) m []

let get_annots ectx seg =
	get_sections_by_name
		name_of_section_for_annots
		(Util.NativeintMap.empty : all_annots Util.NativeintMap.t)
		read_annotations_wrap
		Util.NativeintMap.add
		ectx
		seg
