(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Machine
open Instructions
open Pseudo_insts
open Types
open Types_op
open Integer_op
open Typing_op
open Typing_error
open Output_types

(*
 *
 * Routines for stacks
 *
 *)

exception Invalid_stack_pointer of integer
let failwith_invalid_stack_pointer sp = raise (Invalid_stack_pointer sp)


let expand_stack_stk stk lst =
	List.fold_left (fun stk wt ->
		StkCons (wt, stk)
	) stk lst

let expand_stack_ctx ctx sp lst lst_size new_sp new_freelen =
try
	let cstrt = ctx.ctx_cstrt in
	let str = cstrt.c_store in
	let ics = cstrt.c_ics in

	let ht = IntMap.find sp str.str_locs in
	let rstk = Cast.heap_to_rich_stack ht in

	let freelen = rstk.rstk_freelen in

	let freelen_e = ie_inte freelen in
	let new_freelen_e = ie_inte new_freelen in
	let lst_size_e = ie_inte (IConst lst_size) in

	let ic = freelen_e >=.. lst_size_e in
	if not (Constraint_solver.implies_one ics ic) then
		(* stack overflow *)
		failwith_invalid_stack_pointer sp
	else

	let ie1 = ie_inte freelen in
	let ie2 = new_freelen_e +.. lst_size_e in
	let ic = ie1 =.. ie2 in
	let ics = add_integer_cstrt ic ics in

	let stk = expand_stack_stk rstk.rstk_stk lst in

	let ht = Stack { rstk_stk = stk; rstk_freelen = new_freelen; } in

	(* remove the old stack *)
	let str = remove_store_location sp str in
	(* add the new stack *)
	let str = add_store_location new_sp ht str in

	let cstrt = { cstrt with c_ics = ics; c_store = str; } in
	let ctx = { ctx with ctx_cstrt = cstrt; } in
	ctx
with Not_found ->
	failwith_invalid_stack_pointer sp
| Cast.Error _ ->
	failwith_invalid_stack_pointer sp

let expand_stack ctx sp lst new_sp new_free =

	let num = List.length lst in
	let sz = num * size_of_word in
(*
	let sz = Nativeint.of_int sz in
*)
	let sz = Big_int.big_int_of_int sz in

	let ctx = expand_stack_ctx ctx sp lst sz new_sp new_free in

	let cstrt = ctx.ctx_cstrt in

	let ie1 = ie_inte sp in
	let ie2 = (ie_inte new_sp) +.. (ie_inte (IConst sz)) in
	let ic = ie1 =.. ie2 in
	let ics = add_integer_cstrt ic cstrt.c_ics in

	let cstrt = { cstrt with c_ics = ics; } in
	let ctx = { ctx with ctx_cstrt = cstrt; } in

	ctx

let rec pop_stack_stk_aux stk n poped =
	if n = 0 then
		poped, stk
	else

	match stk with
	| StkCons (wt, stk) ->
		pop_stack_stk_aux stk (n - 1) (wt :: poped)
	| _ ->
		failwithf_type_error "Stack underflow."

let pop_stack_stk stk n =
	let lst, stk = pop_stack_stk_aux stk n [] in
	(List.rev lst, stk)

let pop_stack_ctx ctx sp n new_sp new_freelen =
try
	let cstrt = ctx.ctx_cstrt in
	let str = cstrt.c_store in

	let ht = IntMap.find sp str.str_locs in
	let rstk = Cast.heap_to_rich_stack ht in

	let stk = rstk.rstk_stk in
	let freelen = rstk.rstk_freelen in

	let lst, stk = pop_stack_stk stk n in

	let num = List.length lst in
	let sz = num * size_of_word in
(*
	let sz = Nativeint.of_int sz in
*)
	let sz = Big_int.big_int_of_int sz in

	let ics = cstrt.c_ics in

	let ie1 = ie_inte new_freelen in
	let ie2 = (ie_inte freelen) +.. (ie_inte (IConst sz)) in
	let ic = ie1 =.. ie2 in
	let ics = add_integer_cstrt ic ics in
	let rstk = { rstk_stk = stk; rstk_freelen = new_freelen; } in
	let ht = Stack rstk in

	(* remove the old stack *)
	let str = remove_store_location sp str in
	(* add the new stack *)
	let str = add_store_location new_sp ht str in

	let cstrt = { cstrt with c_ics = ics; c_store = str; } in
	let ctx = { ctx with ctx_cstrt = cstrt; } in
	lst, sz, ctx
with Not_found ->
	failwith_invalid_stack_pointer sp
| Cast.Error _ ->
	failwith_invalid_stack_pointer sp

let pop_stack ctx sp n new_sp new_free =

	let lst, lst_sz, ctx = pop_stack_ctx ctx sp n new_sp new_free in

	let cstrt = ctx.ctx_cstrt in
	let store = cstrt.c_store in

	let ie1 = ie_inte new_sp in
	let ie2 = (ie_inte sp) +.. (ie_inte (IConst lst_sz)) in
	let ic = ie1 =.. ie2 in
	let ics = add_integer_cstrt ic cstrt.c_ics in

	let cstrt = { cstrt with c_ics = ics; c_store = store; } in
	let ctx = { ctx with ctx_cstrt = cstrt; } in

	lst, ctx

(*
let get_two_tvs ctx dnames =
	let dbg_new_sp_name, dnames = Util.pop_with_def "__" dnames in
	let dbg_new_free_name, dnames = Util.pop_with_def "__" dnames in

	let new_tvs, ctx = alloc_inte_tvs [dbg_new_sp_name; dbg_new_free_name] ctx in

	let new_sp = List.nth new_tvs 0 in
	let new_free = List.nth new_tvs 1 in

	(new_sp, new_free), ctx
*)
let get_two_tvs ctx dnames =
	let new_tvs, ctx = alloc_inte_tvs_with_dnames dnames 2 ctx in

	let new_sp = List.nth new_tvs 0 in
	let new_free = List.nth new_tvs 1 in

	(new_sp, new_free), ctx

let push_things dnames ctx lst =
	let (new_sp, new_free), ctx = get_two_tvs ctx dnames in

	(* Don't forget to shift "lst" *)
	let lst = List.map (shift_integer 0 2) lst in

	let esp = P_Op_Reg (Reg32 ESP) in
	let sp = Typing_op.get_integer ctx esp in
	try
		let ctx = expand_stack ctx sp lst new_sp new_free in
		let ctx = Typing_op.set_integer ctx new_sp esp in
		ctx
	with Invalid_stack_pointer sp ->
 		failwithf_type_error "Invalid Stack Pointer: %a\n" Output_types.output_inte (ctx, sp)

let pop dnames ctx n =
	let (new_sp, new_free), ctx = get_two_tvs ctx dnames in

	let esp = P_Op_Reg (Reg32 ESP) in
	let sp = Typing_op.get_integer ctx esp in
	try
		let lst, ctx = pop_stack ctx sp n new_sp new_free in
		let ctx = Typing_op.set_integer ctx new_sp esp in
		lst, ctx
	with Invalid_stack_pointer sp ->
 		failwithf_type_error "Invalid Stack Pointer: %a\n" Output_types.output_inte (ctx, sp)
