(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Types
open Typing_error
open Annotations

open Store_error

open Integer_op

open Util

open Big_int
open Big_int_util



let get_atom ics str base =
try
	let locs = str.str_locs in
	let base' = Types_op.find_location ics base locs in
	base', (IntMap.find base' locs)
with Not_found ->
	failwith_unknown_location base

let get_one_te = function
	| [] ->
		failwithf_type_error "Out of bounds memory access."
	| te :: _ ->
		te

let sub_by_opsize off sz =
	let i = match sz with
		| Instructions.OpSize8 -> 1
		| Instructions.OpSize16 -> 2
		| Instructions.OpSize32 -> 4
		| Instructions.OpSize48 -> 6
		| Instructions.OpSize64 -> 8
		| Instructions.OpSize128 -> 16
	in
	off -^ (bi i)

let rec get_from_tt tt off sz' =
	if off <^ bi_0 then
		failwithf_type_error "Memory access was not aligned."
	else

	match tt with
	| [] ->
		failwithf_type_error "Out of bounds memory access."
	| (sz, inte) :: tt ->

	if off =^ bi_0 then
		if sz = sz' then
			inte
		else
			failwithf_type_error "Memory access was not aligned."
	else
		let off = sub_by_opsize off sz in
		get_from_tt tt off sz'

let rec get_from_stk st off =
	if off <^ bi_0 then
		failwithf_type_error "Non-aligned memory access."
	else

	match st with
	| StkCons (inte, st) ->
		if off =^ bi_0 then
			inte
		else
			let off = off -^ bi_size_of_word in
			get_from_stk st off
	| StkNull ->
		failwithf_type_error "Stack underflow(get)."
	| StkVar _ ->
		failwithf_type_error "Cannot investigate stack because it is abstract(get)."

let get_from_rstk rstk ad sz =
	if sz <> Instructions.OpSize32 then
		failwithf_type_error "Non-aligned stack access has not been supported."
	else
	get_from_stk rstk.rstk_stk ad
	(* ??? : Do we need to check rstk_free? *)

let get_from_ct ics ct off sz =
	match ct with
	| Pack _ ->
		failwith_invalid_location_type (Concrete ct)
	| Tuple tt ->
		get_from_tt tt off sz

let get_from_typ ics typ off sz =
	match typ with
	| Concrete ct ->
		get_from_ct ics ct off sz
	| _ ->
		failwithf_type_error "Load: location's type is not real type"

let get_from_at ics str at off sz =
try
	let typ = Cast.array_type_to_typ str ics at in
	get_from_typ ics typ off sz
with Cast.Error _ ->
	failwith_invalid_location_array_type at

let get ics str ad sz =
	let _, ht = get_atom ics str ad.ad_base in
	match ht with
	| Array at ->
		get_from_at ics str at ad.ad_offset sz
	| Stack rstk ->
		get_from_rstk rstk ad.ad_offset sz



let rec set_to_tt tt off sz' inte =
	if off <^ bi_0 then
		failwithf_type_error "Memory access was not aligned."
	else

	match tt with
	| [] ->
		failwithf_type_error "Out of bounds memory access."
	| te :: tt ->

	let (sz, _) = te in

	if off =^ bi_0 then
		if sz = sz' then
			(sz, inte) :: tt
		else
			failwithf_type_error "Memory access was not aligned."
	else
		let off = sub_by_opsize off sz in
		let tt = set_to_tt tt off sz' inte in
		te :: tt

let rec set_to_stk st off inte =
	if off <^ bi_0 then
		failwithf_type_error "Memory access was not aligned."
	else

	match st with
	| StkCons (inte', st) ->

		if off =^ bi_0 then
			StkCons (inte, st)
		else
			let off = off -^ bi_size_of_word in
			let st = set_to_stk st off inte in
			StkCons (inte', st)

	| StkNull ->
		failwithf_type_error "Stack underflow(set)."
	| StkVar _ ->
		failwithf_type_error "Cannot investigate stack because it is abstract(set)."

let set_to_rstk ics rstk off sz inte =
	if sz <> Instructions.OpSize32 then
		failwithf_type_error "Non-aligned stack access has not been supported."
	else
	let stk = set_to_stk rstk.rstk_stk off inte in
	{ rstk with rstk_stk = stk; }

let set_to_ct ct off sz inte =
	match ct with
	| Pack _ ->
		failwith_invalid_location_type (Concrete ct)
	| Tuple tt ->
		let tt = set_to_tt tt off sz inte in
		Tuple tt

let set_to_typ typ off sz inte =
	match typ with
	| Concrete ct ->
		let ct = set_to_ct ct off sz inte in
		Concrete ct
	| _ ->
		failwith_invalid_location_type typ

let set_to_at ics str at off sz inte =
try
	let t = Cast.array_type_to_typ str ics at in
	let t = set_to_typ t off sz inte in
	{ at with at_type = t; }
with Cast.Error _ ->
	failwith_invalid_location_array_type at

let set ics str inte ad sz =
	let base', ht = get_atom ics str ad.ad_base in
	let off = ad.ad_offset in
	let ht = match ht with
	| Array at ->
		let at = set_to_at ics str at off sz inte in
		Array at
	| Stack rstk ->
		let rstk = set_to_rstk ics rstk off sz inte in
		Stack rstk
	in

	Types_op.replace_store_location base' ht str



let is_in ics str inte =
try
	let _ = Types_op.find_location ics inte str.str_locs in
	true
with Not_found ->
	false

let get_by_integer ics str inte =
	let _, ht = get_atom ics str inte in
	ht

let remove_by_integer ics str inte =
try
	let inte', ht = get_atom ics str inte in
	let locs = IntMap.remove inte' str.str_locs in
	let str = { str with str_locs = locs; } in
	ht, str
with Not_found ->
	failwith_unknown_location inte

let replace_by_integer ics str inte ht =
try
	let locs = str.str_locs in
	let inte' = Types_op.find_location ics inte locs in
	let locs = IntMap.remove inte' locs in
	let locs = IntMap.add inte ht locs in
	let str = { str with str_locs = locs; } in
	str
with Not_found ->
	failwith_unknown_location inte



(*****************************************
	Routines for Code Store
 *****************************************)
let is_in_code_store ics c_str inte =
try
	let _ = Types_op.find_location ics inte c_str in
	true
with Not_found ->
	false

let get_by_integer_in_code_store ics c_str inte =
try
	let inte' = Types_op.find_location ics inte c_str in
	IntMap.find inte' c_str
with Not_found ->
	failwith_unknown_location inte

let add_by_integer_in_code_store ics c_str inte ct =
try
	let inte' = Types_op.find_location ics inte c_str in
	let c_str = IntMap.remove inte' c_str in
	IntMap.add inte ct c_str
with Not_found ->
	IntMap.add inte ct c_str
