(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Machine
open Instructions
open Pseudo_insts
open Disassemble
open Types
open Types_op
open Typing_common
open Typing_error
open Typing_insts
open Output_types
open Util
open Big_int_util

(*
 * Null type-checker state for your convenience.
 *)

let zero_reg32 = RK_E, (IConst bi_0)

let null_ctx =
{
	ctx_kinds = [];
	ctx_cstrt = { c_ics = empty_integer_cstrts; c_store = empty_store; c_code_store = empty_code_store; };
	ctx_registers =
	{
		cs = CodeSeg(SegSize32);

		eax = zero_reg32;
		ebx = zero_reg32;
		ecx = zero_reg32;
		edx = zero_reg32;
		esi = zero_reg32;
		edi = zero_reg32;
		esp = zero_reg32;
		ebp = zero_reg32;

		flags_opt = None;
	};
}

let get_named_type_definition ntds name =
	try
		StringMap.find name ntds
	with Not_found ->
		raise (Type_error (Printf.sprintf "Type %s is not defined." name))
 
let type_integer_sub inte nword =
	match inte with
	| IConst n ->
(*
		(if n <> nword then 
		error_integer_mismatch n nword)
*)
		let nword_signed = big_int_of_nativeint_signed nword in
		let nword_unsigned = big_int_of_nativeint_unsigned nword in
		if n <>^ nword_signed && n <>^ nword_unsigned then
		error_integer_mismatch n nword
	| IVar tv ->
		failwithf_type_error "integer cannot be a type variable."

let get_getter = function
	| OpSize8 -> get_uint8
	| OpSize16 -> get_uint16
	| OpSize32 -> get_uint32
	| _ -> failwithf_type_error "Non supported integer size."

let type_integer seg off inte sz =
	let getter = get_getter sz in
	let nword = getter seg off in
	type_integer_sub inte nword

let add_by_opsize off sz =
	let i = match sz with
		| OpSize8 -> 1
		| OpSize16 -> 2
		| OpSize32 -> 4
		| OpSize48 -> 6
		| OpSize64 -> 8
		| OpSize128 -> 16
	in
	off + i

let rec type_tuple_aux tt seg off =
	match tt with
	| [] ->
		()
	| (sz, inte) :: tt ->
	begin
		type_integer seg off inte sz;
		let off = add_by_opsize off sz in
		type_tuple_aux tt seg off
	end

let type_tuple tt seg off =
	let sz = size_of_tuple_type tt in
	if off + sz <> seg.seg_size then
		error_tuple_size_mismatch sz seg.seg_size
	else
		type_tuple_aux tt seg off
(*
	let tuple_size = (List.length tt) * size_of_word in

	if off + tuple_size <> seg.seg_size
	then
		error_tuple_size_mismatch tuple_size seg.seg_size
	else
		for i = 0 to tuple_size - 1 do
			let off' = off + (i * size_of_word) in
			let inte = List.nth tt i in
			type_integer seg off' inte
		done
*)

let type_state gctx state seg =
	(* code size = 0 indicates that the code is 'trusted' *)
	if seg.seg_size = 0 then
		()
	else

	let (op_attr, ad_attr) =
		match state.st_registers.cs with
		| CodeSeg sz ->
			(
			match sz with
			| SegSize32 -> (OpSizeAttr32, AdSizeAttr32)
			| _ -> failwithf "type code doesn't support 64 bit code"
			)
		| _ -> failwithf "SHOULD NOT OCCUR: cs must be code segment!" in
	let insts = disassemble op_attr ad_attr seg in

	type_insts gctx state seg.seg_va insts

let rec type_stack st seg off =
	let seg_size = seg.seg_size in
	match st with
	| StkVar (tv) ->
		failwithf_type_error "Stack cannot be a type variable."
	| StkNull ->
		if (seg_size <> off) then error_stack_not_null (seg_size - off)
	| StkCons (inte, st') ->
		if (seg_size < off + size_of_word) then error_stack_too_short (off + size_of_word - seg_size)
		type_integer seg off inte;
		type_stack st' seg (off + size_of_word)

let type_rich_stack rstk seg =
	let freelen = rstk.rstk_freelen in

	match freelen with
	| IConst fr ->
(*
		type_stack rstk.rstk_stk seg (Nativeint.to_int fr)
*)
		type_stack rstk.rstk_stk seg (Big_int.int_of_big_int fr)
	| _ ->
		failwithf_type_error "Stack type should have a concrete free size."

let type_concrete_type ct seg off =
	match ct with
	| Pack _ ->
		error_incompatible_concrete_type null_ctx ct
	| Tuple tt ->
		type_tuple tt seg off

let type_name gctx name cs seg off =
	let ntd = get_named_type_definition gctx.gctx_named_type_definitions name in
	let bdy = apply_named_type_definition ntd cs in
	match bdy with
	| NtdConcrete ct ->
		type_concrete_type ct seg off
	| _ ->
		failwithf_type_error "@[%s is not a data type.@]" name

let type_name_code gctx name cs seg =
	let ntd = get_named_type_definition gctx.gctx_named_type_definitions name in
	let bdy = apply_named_type_definition ntd cs in
	match bdy with
	| NtdState st ->
		type_state gctx st seg
	| _ ->
		failwithf_type_error "@[%s is not a code type.@]" name

let type_code gctx ct seg =
	match ct with
	| NameCode (name, cs) ->
		type_name_code gctx name cs seg
	| State st ->
		type_state gctx st seg

let type_typ gctx typ seg off =
	match typ with
	| Concrete (concrete_type) ->
		type_concrete_type concrete_type seg off
	| Name (name, cs) ->
		type_name gctx name cs seg off

let type_array gctx at seg =
	let typ = at.at_type in
	let len = at.at_len in

	let sizeof_typ = size_of_type typ in
	match len with
(*
	| IConst sz when seg.seg_size = (Nativeint.to_int sz) * sizeof_typ ->
*)
	| IConst sz when seg.seg_size = (Big_int.int_of_big_int sz) * sizeof_typ ->
		for i = 0 to (Big_int.int_of_big_int sz) do
			type_typ gctx typ seg (i * sizeof_typ)
		done
	| _ ->
		failwithf_type_error "@[Segment is not fully occupied.@]"

let type_heap gctx ht seg =
	match ht with
	| Array at ->
		type_array gctx at seg
	| Stack rstk ->
		type_rich_stack rstk seg

let type_block gctx blk =
	let seg =  blk.blk_segment in
	let ht = blk.blk_heap_type in

	Format.printf "@[Begin typing 0x%08nx..@]@." seg.seg_va;

	try
		type_heap gctx ht seg;
		Format.printf "@[Done.@]@."

	with Type_error (mesg) ->
		error_at_typing_block mesg blk

let type_code_block gctx c_blk =
	let seg = c_blk.cblk_segment in
	let ct = c_blk.cblk_code_type in

	Format.printf "@[Begin typing 0x%08nx..@]@." seg.seg_va;
	try
		type_code gctx ct seg;
		Format.printf "@[Done.@]@."
	with Type_error (mesg) ->
		error_at_typing_code_block mesg c_blk	

let is_consistent_cstrt cstrt =
	let ics = cstrt.c_ics in
	let str = cstrt.c_store in
(* XXX : does not check anything!
	Constraint_solver.is_consistent str [] ics
*)
	true

let is_consistent_st st =
	let cstrts = st.st_prefix.p_cstrts in
	List.for_all is_consistent_cstrt cstrts

let is_consistent_ct = function
	| State st ->
		is_consistent_st st
	| _ ->
		true

let type_check name_to_named_type_definitions va_to_block va_to_code_block va_to_annots =
	let code_store = NativeintMap.fold (fun n c_blk c_str ->
		let ct = c_blk.cblk_code_type in
		if is_consistent_ct ct then
			let n = big_int_of_nativeint_unsigned n in
			add_code_location (IConst n) ct c_str
		else
			failwithf_type_error "@[@]@."
	) va_to_code_block empty_code_store in
	(*
	 * XXX : it would be nice to create "store" here and use it as the initial memory for the entry function.
	 * In the case, do not forget to shift stacks for free-size.
	 *)
	let gctx =
	{
		gctx_named_type_definitions	= name_to_named_type_definitions;
		gctx_code_store			= code_store;
		gctx_annots_map			= va_to_annots;
	} in
	NativeintMap.iter (fun address block -> type_block gctx block) va_to_block;
	NativeintMap.iter (fun address c_blk -> type_code_block gctx c_blk) va_to_code_block
