(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Types
open Types_op
open Typing_op
open Integer_op
open Annotations
open Pseudo_insts
open Typing_common
open Typing_error

open Util
open Big_int_util

open Instructions

(*
 *
 * Routines for typing pseudo instructions.
 *
 *)

let add_cstrt_if_consistent cstrt1 cstrt2 result =
try
	(*
	 * It's ok NOT to check wheather if str1 and str2 are consistend,
	 * because str1 and str2 are separated by the type system.
	 *)
	let store = add_store cstrt1.c_store cstrt2.c_store in
	let code_store = add_code_store cstrt1.c_code_store cstrt2.c_code_store in

	let ics1 = cstrt1.c_ics in
	let ics2 = cstrt2.c_ics in

(* XXX : Add always! *)
(*
	if not (Constraint_solver.is_consistent store ics1 ics2) then
		result
	else
*)

	let ics = add_integer_cstrts ics1 ics2 in

	let cstrt = { c_ics = ics; c_store = store; c_code_store = code_store; } in
	cstrt :: result
with Invalid_store_manipulation ->
	result

let open_package ctx pt names =
	let kinds1 = ctx.ctx_kinds in
	let cstrt1 = ctx.ctx_cstrt in
	let regs1 = ctx.ctx_registers in

	let prefix2 = pt.pt_prefix in
	let kinds2 = prefix2.p_kinds in
	let cstrts2 = prefix2.p_cstrts in

	let kinds2 = List.map (fun (k, _) -> k) kinds2 in
	let kinds2 =
	try
		List.combine kinds2 names
	with Invalid_argument _ ->
		failwithf_type_error "@[Too few names for new type variables.@]" 
	in

	let n = List.length kinds2 in

	let cstrt1 = shift_cstrt 0 n cstrt1 in
	let regs1 = shift_registers_type 0 n regs1 in

	let kinds = kinds2 @ kinds1 in

	let cstrts =
		if cstrts2 = [] then
			[cstrt1]
		else
			List.fold_right (add_cstrt_if_consistent cstrt1) cstrts2 []
	in

	(* ??? : is this check necessary? : if cstrts = [] ?*)
	match cstrts with
	| [] ->
		failwithf_type_error "@[%a@ cannot be unpacked.@]" 
			Output_types.output_pt (ctx, pt)
	| _ ->

	let prefix = { p_kinds = kinds; p_cstrts = cstrts; } in
	let state = { st_prefix = prefix; st_registers = regs1; } in

	let tt = pt.pt_tuple in

	state, tt, n


let add_type_cstrt cstrt inte typ =
	let store = cstrt.c_store in
	let store = add_store_location inte typ store in

	let cstrt = { cstrt with c_store = store; } in
	cstrt

let add_type ctx inte typ =
	let cstrt = add_type_cstrt ctx.ctx_cstrt inte typ in
	let ctx = { ctx with ctx_cstrt = cstrt; } in
	ctx

let add_type_to_prefix prefix inte typ =
	let cstrts = List.map (fun cstrt -> add_type_cstrt cstrt inte typ) prefix.p_cstrts in
	let prefix = { prefix with p_cstrts = cstrts; } in
	prefix

let add_type_to_state state inte typ =
	let prefix = add_type_to_prefix state.st_prefix inte typ in
	let state = { state with st_prefix = prefix; } in
	state

let acc_if_implies ctx cstrt1 cstrt2 result =
	let { c_ics = ics1; c_store = str1; c_code_store = c_str1; } = cstrt1 in
	let { c_ics = ics2; c_store = str2; c_code_store = c_str2; } = cstrt2 in


	if not (Constraint_solver.implies ics1 ics2) then
(
(*
	Format.printf "@[%a@]@." Output_types.output_inte_cstrts (ctx, ics1);
	Format.printf "@[===>@]@.";
	Format.printf "@[%a@]@.@." Output_types.output_inte_cstrts (ctx, ics2);
*)
		result
)
	else

try
	let str =

try
		sub_store ics1 str1 str2
with Invalid_store_manipulation ->
(*
	Format.printf "@[%a@]@." Output_types.output_store (ctx, str1);
	Format.printf "@[===>@]@.";
	Format.printf "@[%a@]@.@." Output_types.output_store (ctx, str2);
*)
	raise Invalid_store_manipulation
	in

	let _ =
try
		sub_code_store ics1 c_str1 c_str2
with Invalid_store_manipulation ->
(*
	Format.printf "@[%a@]@." Output_types.output_code_store (ctx, c_str1);
	Format.printf "@[===>@]@.";
	Format.printf "@[%a@]@.@." Output_types.output_code_store (ctx, c_str2);
*)
	raise Invalid_store_manipulation
	in


	let cstrt = { cstrt1 with c_store = str; c_code_store = c_str1; } in
	cstrt :: result
with Invalid_store_manipulation ->
	result

let have_same_stores ics = function
	| [] ->
		true
	| cstrt :: cstrts ->
		let str = cstrt.c_store in
		let strs = List.map (fun cstrt -> cstrt.c_store) cstrts in
		List.for_all (is_equal_store ics str) strs

let type_pinst_PPACK ctx inte pt cs =
try
	let ht1, ctx = remove_from_context ctx inte in
	let at1 = Cast.heap_to_array ht1 in

	let cstrt1 = ctx.ctx_cstrt in
	let ics1 = cstrt1.c_ics in
	let str1 = cstrt1.c_store in

	let typ1 = Cast.array_type_to_typ str1 ics1 at1 in

(*
	Format.printf "@[cs = %a@]@." Output_types.output_cs (ctx, cs);
*)

	let (cstrts2, tt) = Types_op.apply_pack_type pt cs in
	let typ2 = Concrete (Tuple tt) in

	if not (Types_op.is_equal_type ics1 typ1 typ2) then
		failwithf_type_error "@[%a@ cannot be packed into@ %a. (%a)@]" 
			Output_types.output_type (ctx, typ1)
			Output_types.output_pt (ctx, pt)
			Output_types.output_type (ctx, typ2)
	else

	let at = { at1 with at_type = Concrete (Pack pt); } in
	let ht = Array at in
	let ctx = add_type ctx inte ht in

	let cstrt1 = ctx.ctx_cstrt in
	let ics1 = cstrt1.c_ics in
	let str1 = cstrt1.c_store in

	let cstrts = 
		if cstrts2 = [] then
			[cstrt1]
		else
			List.fold_right (acc_if_implies ctx cstrt1) cstrts2 []
	in

	(* ??? : check whether if all possible contexts have the same store (for preventing memory leak) *)
	if cstrts = [] || not (have_same_stores ics1 cstrts) then
		failwithf_type_error "@[%a@ cannot be packed into@ %a.@]" 
			Output_types.output_type (ctx, typ1)
			Output_types.output_pt (ctx, pt)
	else

	let prefix = { p_kinds = ctx.ctx_kinds; p_cstrts = cstrts; } in
	{ st_prefix = prefix; st_registers = ctx.ctx_registers; }
with Too_many_arguments ->
	failwithf_type_error "@[Too many arguments for %a.@]" Output_types.output_pt (ctx, pt)
| Too_few_arguments ->
	failwithf_type_error "@[Too few arguments for %a.@]" Output_types.output_pt (ctx, pt)
| Invalid_store_manipulation ->
(
(*
	Format.printf "@[%a@]@." Output_types.output_context ctx;
*)
	failwithf_type_error "@[Invalid store manipulation for %a.@]" Output_types.output_pt (ctx, pt)
)
| Cast.Error _ ->
	failwithf_type_error "@[Array cannot be packed into %a.@]" Output_types.output_pt (ctx, pt)

let type_pinst_PUNPACK ctx inte names =
	let names = List.rev names in
try
	let htyp, ctx = remove_from_context ctx inte in
	let atyp = Cast.heap_to_array htyp in

	let ics = ctx.ctx_cstrt.c_ics in
	let str = ctx.ctx_cstrt.c_store in
	let typ = Cast.array_type_to_typ str ics atyp in

	let pt = Cast.concrete_to_package (Cast.typ_to_concrete typ) in

	let state, tt, n = open_package ctx pt names in

	let inte = shift_integer 0 n inte in
	let len = shift_integer 0 n atyp.at_len in
	let at = { at_type = Concrete (Tuple tt); at_len = len; } in
	let ht = Array at in
	let state = add_type_to_state state inte ht in

	state
with Cast.Error _ ->
	failwithf_type_error "%a didn't point to an existential package."
		Output_types.output_inte (ctx, inte)

let get_named_type_definition ntds name =
try
	StringMap.find name ntds
with Not_found ->
	failwithf_type_error "No such type %s." name

let get_ct ics global_c_str c_str inte =
try
	Store.get_by_integer_in_code_store ics c_str inte
with Store_error.Unknown_location _ ->
	Store.get_by_integer_in_code_store ics global_c_str inte

let roll_common getter is_equal replacer m inte n a =
	let t = getter m inte in

	if is_equal t a then
		replacer m inte n
	else
		raise Not_found

let get_type_by_integer ics store inte =
try
	let ht = Store.get_by_integer ics store inte in
	let at = Cast.heap_to_array ht in
	Cast.array_type_to_typ store ics at
with _ ->
	raise Not_found

let replace_type_by_integer ics store inte typ =
	let at = { at_type = typ; at_len = IConst bi_1; } in
	let ht = Array at in
	Store.replace_by_integer ics store inte ht

let roll_in_store ics store inte nt ct =
	roll_common (get_type_by_integer ics) (Types_op.is_equal_type ics) (replace_type_by_integer ics)
		store inte (Name nt) (Concrete ct)

let test ctx ics ct1 ct2 =
	Format.printf "@[%a@]@." Output_types.output_code (ctx, ct1);
	Format.printf "==";
	Format.printf "@[%a@]@." Output_types.output_code (ctx, ct2);
	Types_op.is_equal_code_type ics ct1 ct2

let roll_in_code_store ctx ics global_c_str c_str inte nt st =
	roll_common (get_ct ics global_c_str) (Types_op.is_equal_code_type ics)
(*
	roll_common (get_ct ics global_c_str) (test ctx ics)
*)
		(* NOTICE: "add", not "replace". code location can be imported from global code store *)
		(Store.add_by_integer_in_code_store ics)
		c_str inte (NameCode nt) (State st)

let roll gctx ctx inte nt =

	let ntds = gctx.gctx_named_type_definitions in
	let global_c_str = gctx.gctx_code_store in

	let cstrt = ctx.ctx_cstrt in
	let store = cstrt.c_store in
	let c_str = cstrt.c_code_store in
try
	let (name, cs) = nt in
	let ntd = get_named_type_definition ntds name in
	let bdy = apply_named_type_definition ntd cs in

	let ics = ctx.ctx_cstrt.c_ics in
	let cstrt =
		match bdy with
		| NtdConcrete ct ->
			let store = roll_in_store ics store inte nt ct in
			{ cstrt with c_store = store; }
		| NtdState st ->
			let c_str = roll_in_code_store ctx ics global_c_str c_str inte nt st in
			{ cstrt with c_code_store = c_str; }
	in

	{ ctx with ctx_cstrt = cstrt; }

with Too_many_arguments ->
	failwithf_type_error "@[Too many arguments for %a.@]" Output_types.output_nt (ctx, nt)
| Too_few_arguments ->
	failwithf_type_error "@[Too few arguments for %a.@]" Output_types.output_nt (ctx, nt)
| Not_found ->
	failwithf_type_error "@[%a cannot be rolled into %a.@]"
		Output_types.output_inte (ctx, inte)
		Output_types.output_nt (ctx, nt)

let type_pinst_PROLL gctx ctx inte nt =
	roll gctx ctx inte nt

let get_name_type gctx ctx inte =
	let cstrt = ctx.ctx_cstrt in
	let ics = cstrt.c_ics in
	let store = cstrt.c_store in
	let c_str = cstrt.c_code_store in
	let global_c_str = gctx.gctx_code_store in
try
	if Store.is_in ics store inte then
		let ht = Store.get_by_integer ics store inte in
		let at = Cast.heap_to_array ht in
		let typ = Cast.array_type_to_typ store ics at in
		Cast.typ_to_named_type typ
	else if Store.is_in_code_store ics c_str inte then
		let ct = Store.get_by_integer_in_code_store ics c_str inte in
		Cast.code_to_named_type ct
	else if Store.is_in_code_store ics global_c_str inte then
		let ct = Store.get_by_integer_in_code_store ics global_c_str inte in
		Cast.code_to_named_type ct
	else
		failwithf_type_error "@[%a is not a valid memory address.@]"
		Output_types.output_inte(ctx, inte)
with Cast.Error _ ->
	failwithf_type_error "@[%a didn't point to a named type@ where@ %a.@]"
		Output_types.output_inte (ctx, inte)
		Output_types.output_context ctx

let unroll_in_store ics store inte ct =
	let at = { at_type = Concrete ct; at_len = IConst bi_1; } in
	let ht = Array at in
	Store.replace_by_integer ics store inte ht

let unroll_in_code_store ics (*global_c_str*) c_str inte st =
	Store.add_by_integer_in_code_store ics c_str inte (State st)


let unroll gctx ctx inte =

	let ntds = gctx.gctx_named_type_definitions in
	let global_c_str = gctx.gctx_code_store in

	let (name, cs) = get_name_type gctx ctx inte in
	let cstrt = ctx.ctx_cstrt in
	let store = cstrt.c_store in
	let c_str = cstrt.c_code_store in
	let ics = cstrt.c_ics in
try
	let ntd = get_named_type_definition ntds name in
	let bdy = apply_named_type_definition ntd cs in

	let cstrt =
		match bdy with
		| NtdConcrete ct ->
			let store = unroll_in_store ics store inte ct in
			{ cstrt with c_store = store; }
		| NtdState st ->
			let c_str = unroll_in_code_store ics (*global_c_str*) c_str inte st in
			{ cstrt with c_code_store = c_str; }
	in

	{ ctx with ctx_cstrt = cstrt; }
with Too_many_arguments ->
	failwithf_type_error "Too many arguments for %s." name
| Too_few_arguments ->
	failwithf_type_error "Too few arguments for %s." name
| Store_mismatch str ->
	failwithf_type_error "Store didn't match for %s." name

let type_pinst_PUNROLL ntds ctx inte =
	unroll ntds ctx inte

let type_pinst_PSPLIT dnames ctx orig_loc req_len =
try
	let (new_loc, new_len), ctx = alloc_inte_two_tvs_with_dnames dnames ctx  in
	let orig_loc = shift_integer 0 2 orig_loc in
	let req_len = shift_integer 0 2 req_len in

	let ht1, ctx = remove_from_context ctx orig_loc in
	let at = Cast.heap_to_array ht1 in

	let orig_type = at.at_type in
	let orig_len = at.at_len in

	let cstrt = ctx.ctx_cstrt in
	let str = cstrt.c_store in
	let ics = cstrt.c_ics in

	let ic = (ie_inte orig_len) >.. (ie_inte req_len) in
	if not (Constraint_solver.implies_one ics ic) then
		failwithf_type_error "The array is too short: %a (requested length = %a)"
			Output_types.output_inte (ctx, orig_len)
			Output_types.output_inte (ctx, req_len)
	else

	let ic = (ie_inte req_len) >.. (ie_int 0) in
	if not (Constraint_solver.implies_one ics ic) then
(
(*
Format.printf "@[ics = %a@]@." Output_types.output_inte_cstrts (ctx, ics);	
*)
		failwithf_type_error "The array cannot be split by %a."
			Output_types.output_inte (ctx, req_len)
)
	else

	let elm_sz = size_of_type orig_type in
	let ie0 = (ie_int elm_sz) *.. (ie_inte req_len) in
	let ie1 = (ie_inte orig_loc) +.. ie0 in
	let ie2 = ie_inte new_loc in
	let ic = ie1 =.. ie2 in
	let ics = ic :: ics in

	let ie1 = (ie_inte req_len) +.. (ie_inte new_len) in
	let ie2 = ie_inte orig_len in
	let ic = ie1 =.. ie2 in
	let ics = ic :: ics in

	let cstrt = { cstrt with c_ics = ics; } in
	let ctx = { ctx with ctx_cstrt = cstrt; } in

	let ht_a = Array { at_type = orig_type; at_len = req_len; } in
	let ht_b = Array { at_type = orig_type; at_len = new_len; } in

	let ctx = add_type ctx orig_loc ht_a in
	let ctx = add_type ctx new_loc ht_b in

	ctx
with Cast.Error _ ->
	failwithf_type_error "There is no array in %a." Output_types.output_inte (ctx, orig_loc)

let type_pinst_PCONCAT dnames ctx orig_loc1 orig_loc2 =
try
	let new_len, ctx = alloc_inte_tv_with_dnames dnames ctx  in
	let orig_loc1 = shift_integer 0 1 orig_loc1 in
	let orig_loc2 = shift_integer 0 1 orig_loc2 in

	let ht1, ctx = remove_from_context ctx orig_loc1 in
	let ht2, ctx = remove_from_context ctx orig_loc2 in
	let at1 = Cast.heap_to_array ht1 in
	let at2 = Cast.heap_to_array ht2 in

	let cstrt = ctx.ctx_cstrt in
	let str = cstrt.c_store in
	let ics = cstrt.c_ics in

	if not (is_equal_type ics at1.at_type at2.at_type) then
		failwithf_type_error "Cannot concatenate different type of arrays."
	else

	let orig_type = at1.at_type in
	let orig_len1 = at1.at_len in
	let orig_len2 = at2.at_len in

	let elm_sz = size_of_type orig_type in
	let ie0 = (ie_int elm_sz) *.. (ie_inte orig_len1) in
	let ie1 = (ie_inte orig_loc1) +.. ie0 in
	let ie2 = ie_inte orig_loc2 in
	let ic = ie1 =.. ie2 in
(*
	if not (Constraint_solver.implies str ics [ic]) then
*)
	if not (Constraint_solver.implies_one ics ic) then
		failwithf_type_error "Cannot concatenate non-adjacent arrays"
	else

	let ie1 = (ie_inte orig_len1) +.. (ie_inte orig_len2) in
	let ie2 = ie_inte new_len in
	let ic = ie1 =.. ie2 in
	let ics = ic :: ics in

	let cstrt = { cstrt with c_ics = ics; } in
	let ctx = { ctx with ctx_cstrt = cstrt; } in

	let ht = Array { at_type = orig_type; at_len = new_len; } in

	let ctx = add_type ctx orig_loc1 ht in

	ctx
with Cast.Error _ ->
	failwithf_type_error "There is no array in %a or %a."
		Output_types.output_inte (ctx, orig_loc1)
		Output_types.output_inte (ctx, orig_loc2)

let rec split_tuple_aux tt req_sz =
	if req_sz <= 0 then
		([], tt, 0)
	else

	match tt with
	| [] ->
		failwithf_typechecker_bug "split_tuple: requested size should have been tested..."
	| inte :: tt ->
		let tt1, tt2, sz = split_tuple_aux tt (req_sz - 1) in

		(* currently, the size of a tuple element == size_of_word *)
		(inte :: tt1, tt2, sz + Machine.size_of_word)

let split_tuple tt req_sz =
	if req_sz <= 0 then
		failwithf_type_error "TUPLE_SPLIT: Cannot split by <0."
	else

	let len = List.length tt in
	if len <= req_sz then
		failwithf_type_error "TUPLE_SPLIT: Cannot split by %d." req_sz
	else

	split_tuple_aux tt req_sz

let concat_tuple tt1 tt2 =
	tt1 @ tt2

let type_pinst_PTUPLE_SPLIT dnames ctx orig_loc req_len =
	let req_len =
		match req_len with
		| IConst i ->
(*
			Nativeint.to_int i
*)
			Big_int.int_of_big_int i
		| _ ->
			failwithf_type_error "TUPLE_SPLIT: Cannot split tuple by unknown size %a."
					Output_types.output_inte (ctx, req_len)
	in
try
	let new_loc, ctx = alloc_inte_tv_with_dnames dnames ctx in
	let orig_loc = shift_integer 0 1 orig_loc in

	let ht1, ctx = remove_from_context ctx orig_loc in
	let at = Cast.heap_to_array ht1 in

	let cstrt = ctx.ctx_cstrt in
	let str = cstrt.c_store in
	let ics = cstrt.c_ics in

	let ic = (ie_inte at.at_len) =.. (ie_int 1) in
(*
	if not (Constraint_solver.implies str ics [ic]) then
*)
	if not (Constraint_solver.implies_one ics ic) then
		failwithf_type_error "TUPLE_SPLIT: the size of the array must be 1."
	else

	let tt = Cast.concrete_to_tuple (Cast.typ_to_concrete at.at_type) in
	let tt1, tt2, sz = split_tuple tt req_len in
	let t1 = Concrete (Tuple tt1) in
	let t2 = Concrete (Tuple tt2) in

	let ie1 = (ie_inte orig_loc) +.. (ie_int sz) in
	let ie2 = ie_inte new_loc in
	let ic = ie1 =.. ie2 in
	let ics = ic :: ics in

	let cstrt = { cstrt with c_ics = ics; } in
	let ctx = { ctx with ctx_cstrt = cstrt; } in

	let ht_a = Array { at_type = t1; at_len = IConst bi_1; } in
	let ht_b = Array { at_type = t2; at_len = IConst bi_1; } in

	let ctx = add_type ctx orig_loc ht_a in
	let ctx = add_type ctx new_loc ht_b in

	ctx
with Cast.Error _ ->
	failwithf_type_error "There is no array in %a." Output_types.output_inte (ctx, orig_loc)

let type_pinst_PTUPLE_CONCAT ctx orig_loc1 orig_loc2 =
try
	let ht1, ctx = remove_from_context ctx orig_loc1 in
	let ht2, ctx = remove_from_context ctx orig_loc2 in
	let at1 = Cast.heap_to_array ht1 in
	let at2 = Cast.heap_to_array ht2 in

	let cstrt = ctx.ctx_cstrt in
	let str = cstrt.c_store in
	let ics = cstrt.c_ics in

	let ic = (ie_inte at1.at_len) =.. (ie_int 1) in
(*
	if not (Constraint_solver.implies str ics [ic]) then
*)
	if not (Constraint_solver.implies_one ics ic) then
		failwithf_type_error "TUPLE_CONCAT: the size of the array must be 1."
	else
	let ic = (ie_inte at2.at_len) =.. (ie_int 1) in
(*
	if not (Constraint_solver.implies str ics [ic]) then
*)
	if not (Constraint_solver.implies_one ics ic) then
		failwithf_type_error "TUPLE_CONCAT: the size of the array must be 1."
	else

	let tt1 = Cast.concrete_to_tuple (Cast.typ_to_concrete at1.at_type) in
	let tt2 = Cast.concrete_to_tuple (Cast.typ_to_concrete at2.at_type) in

	let sz = size_of_tuple_type tt1 in

	let ie1 = (ie_inte orig_loc1) +.. (ie_int sz) in
	let ie2 = ie_inte orig_loc2 in
	let ic = ie1 =.. ie2 in
(*
	if not (Constraint_solver.implies str ics [ic]) then
*)
	if not (Constraint_solver.implies_one ics ic) then
		failwithf_type_error "TUPLE_CONCAT: the tuples are not adjacent."
	else

	let tt = concat_tuple tt1 tt2 in
	let t = Concrete (Tuple tt) in
	let ht = Array { at_type = t; at_len = IConst bi_1; } in

	let ctx = add_type ctx orig_loc1 ht in

	ctx
with Cast.Error _ ->
	failwithf_type_error "There is no array in %a or %a."
		Output_types.output_inte (ctx, orig_loc1)
		Output_types.output_inte (ctx, orig_loc2)


let remove_if_exist ics inte store =
try
	let _, store = Store.remove_by_integer ics store inte in
	store
with Store_error.Unknown_location _ ->
	store

let type_jumps global_c_str ctx inte =
try
	let ct = get_ct ctx.ctx_cstrt.c_ics global_c_str ctx.ctx_cstrt.c_code_store inte in

	let state = Cast.code_to_state ct in

	if Unify.is_subtype_state_type global_c_str ctx state then
		ctx
	else
		failwithf_type_error "@[Code type@ %a@ cannot be instantiated to the current context@ %a@]"
					Output_types.output_code (ctx, ct)
					Output_types.output_context ctx
with Store_error.Unknown_location inte ->
	failwithf_type_error "%a is not a code location." Output_types.output_inte (ctx, inte)
| Cast.Error et ->
begin
	match et with
	| Cast.Cast_err_code ct ->
		failwithf_type_error "%a wasn't a code type." Output_types.output_code (ctx, ct)
	| _ ->
		failwithf_typechecker_bug "Impossible Cast Error was raised."
end

let adjust_relative_ad nva op =
	match op with
	| P_Op_Imd (n, sz) ->
		let nva = big_int_of_nativeint_unsigned nva in
		let n = nva +^ n in
		P_Op_Imd (n, sz)
	| _ ->
		op

let type_pinst_JMPN global_c_str ctx nva op =
(*
	let op = match op with
		| P_Op_Imd n ->
(*
			let n = Nativeint.add nva n in
*)
			let nva = big_int_of_nativeint_unsigned nva in
			let n = nva +^ n in
			P_Op_Imd n
		| _ ->
			op in
*)
	let op = adjust_relative_ad nva op in
	let inte = Typing_op.get_integer ctx op in
	type_jumps global_c_str ctx inte


let type_pinst_CALLN global_c_str ctx nva op =
	let op = adjust_relative_ad nva op in
	let inte = Typing_op.get_integer ctx op in

	let dnames = ["__CALL_SP"; "__CALL_FREE";] in
	let nva = big_int_of_nativeint_unsigned nva in
	let ctx = Stack.push_things dnames ctx [(IConst nva)] in
	type_jumps global_c_str ctx inte

let type_pinst_RET global_c_str ctx =
	let dnames = ["__RET_SP"; "__RET_FREE"; ] in
	let lst, ctx = Stack.pop dnames ctx 1 in

	if List.length lst <> 1 then
		failwithf_typechecker_bug "One word must be poped from the stack."
	else

	let inte = List.nth lst 0 in
	type_jumps global_c_str ctx inte

let type_pinst_FALL global_c_str ctx nva =
	let nva = big_int_of_nativeint_unsigned nva in
	let inte = IConst nva in
	type_jumps global_c_str ctx inte

let get_branch_condition ctx cc result op1 op2 =
	let result = ie_inte result in
	let zero = ie_int 0 in
	match cc with
	| CC_Z ->
		(result =.. zero), (result <>.. zero)
	| CC_NZ ->
		(result <>.. zero), (result =.. zero)
	| CC_ULT | CC_SLT ->
		(result <.. zero), (result >=.. zero)
	| CC_ULE | CC_SLE ->
		(result <=.. zero), (result >.. zero)
	| CC_UGT | CC_SGT ->
		(result >.. zero), (result <=.. zero)
	| CC_UGE | CC_SGE ->
		(result >=.. zero), (result <.. zero)
	| _ ->
		failwithf_type_error "Unsupported condition code."

exception Skip

let type_pinst_Jcc global_c_str ctx nva cc op =
	let op = adjust_relative_ad nva op in
	let inte = Typing_op.get_integer ctx op in

	let flags_opt = ctx.ctx_registers.flags_opt in
	match flags_opt with
	| None ->
		let _ = type_jumps global_c_str ctx inte in
		ctx
	| Some flags ->

	let result = flags.f_rst in
	let op1 = flags.f_op1 in
	let op2 = flags.f_op2 in

	let ic_taken, ic_not_taken = get_branch_condition ctx cc result op1 op2 in

	let cstrt = ctx.ctx_cstrt in
	let ics = cstrt.c_ics in

	(* branch taken *)
	let ics' = ic_taken :: ics in
	if Constraint_solver.is_satisfiable ics' then
		let cstrt' = { cstrt with c_ics = ics'; } in
		let ctx' = { ctx with ctx_cstrt = cstrt'; } in
		let _ = type_jumps global_c_str ctx' inte in
		()
	else
		();

	(* branch not taken *)
	let ics = ic_not_taken :: ics in
	if not (Constraint_solver.is_satisfiable ics) then
		raise Skip
	else

	let cstrt = { cstrt with c_ics = ics; } in
	let ctx = { ctx with ctx_cstrt = cstrt; } in
	ctx

let type_pinst_MOV ctx op1 op2 =
	let inte = get_integer ctx op2 in
	set_integer ctx inte op1

let rec get_new_names n =
	if n <= 0 then []
	else

	let i = get_counter_value () in
	let name = "_N" ^ (string_of_int i) in
	let names = get_new_names (n - 1) in
	(KInt, name) :: names

let rec get_new_type_variables ctx n =
	let new_names = get_new_names n in
	alloc_tvs ctx new_names

let get_new_type_variable ctx =
	let new_tvs, ctx = get_new_type_variables ctx 1 in
	(List.nth new_tvs 0), ctx

let manipulate_stack_pointer dnames ctx i =

	if (i %^ bi_size_of_word) <>^ bi_0 then
		failwithf_type_error "@[Stack pointer should be shifted with multiples of 4.@]"
	else

	let i = i /^ bi_size_of_word in
	let i = Big_int.int_of_big_int i in

	if i < 0 then
		let i = -i in
		let lst, ctx = get_new_type_variables ctx i in
		let lst = List.map (fun tv -> (IVar tv) ) lst in

		Stack.push_things dnames ctx lst
	else
		let _, ctx = Stack.pop dnames ctx i in
		ctx

let set_flags ctx op_size result op1 op2 =
	let regs = ctx.ctx_registers in
	let flags = { f_rst = result; f_op1 = op1; f_op2 = op2; f_size = op_size; } in
	let regs = { regs with flags_opt = Some flags; } in
	{ ctx with ctx_registers = regs; }

let shift_address c d ad =
	let base = ad.ad_base in
	let offset = ad.ad_offset in
	let base = shift_integer c d base in
	{ ad_base = base; ad_offset = offset; }

let shift_operand c d p_op =
	match p_op with
	| P_Op_Mem (mop, ad) ->
		let ad = shift_address c d ad in
		P_Op_Mem (mop, ad)
	| _ ->
		p_op

let type_pinst_ADD dbg_new_names ctx op1 op2 =

	let op_size1 = get_op_size op1 in
	let op_size2 = get_op_size op2 in

	(* XXX : should we check operand size ? *)

	let inte1 = get_integer ctx op1 in
	let inte2 = get_integer ctx op2 in

	(* Treat the stack pointer (ESP) specially *)
	match op1 with
	| P_Op_Reg (Instructions.Reg32 Instructions.ESP) ->
	begin
		match inte2 with
		| IConst i ->
			manipulate_stack_pointer dbg_new_names ctx i
		| _ ->
			failwithf_type_error "@[Stack pointer should be shifted with integer constatns.@]"
	end
	| _ ->

	let inte, ctx = alloc_inte_tv_with_dnames dbg_new_names ctx in

	let inte1 = shift_integer 0 1 inte1 in
	let inte2 = shift_integer 0 1 inte2 in
	let op1 = shift_operand 0 1 op1 in

	let ie1 = ie_inte inte in
	let ie2 = (ie_inte inte1) +.. (ie_inte inte2) in
	let ic = ie1 =.. ie2 in

	let cstrt = ctx.ctx_cstrt in
	let ics = add_integer_cstrt ic cstrt.c_ics in
	let cstrt = { cstrt with c_ics = ics; } in
	let ctx = { ctx with ctx_cstrt = cstrt; } in

	let ctx = set_flags ctx op_size1 inte inte1 inte2 in

	set_integer ctx inte op1

let type_pinst_SUB dbg_new_names ctx op1 op2 =

	let op_size1 = get_op_size op1 in
	let op_size2 = get_op_size op2 in

	(* XXX : should we check operand size ? *)

	let inte1 = get_integer ctx op1 in
	let inte2 = get_integer ctx op2 in

	(* Treat the stack pointer (ESP) specially *)
	match op1 with
	| P_Op_Reg (Instructions.Reg32 Instructions.ESP) ->
	begin
		match inte2 with
		| IConst i ->
			let i = Big_int.minus_big_int i in
			manipulate_stack_pointer dbg_new_names ctx i
		| _ ->
			failwithf_type_error "@[Stack pointer should be shifted with integer constatns.@]"
	end
	| _ ->

	let inte, ctx = alloc_inte_tv_with_dnames dbg_new_names ctx in

	let inte1 = shift_integer 0 1 inte1 in
	let inte2 = shift_integer 0 1 inte2 in
	let op1 = shift_operand 0 1 op1 in

	let ie1 = ie_inte inte1 in
	let ie2 = (ie_inte inte) +.. (ie_inte inte2) in
	let ic = ie1 =.. ie2 in

	let cstrt = ctx.ctx_cstrt in
	let ics = add_integer_cstrt ic cstrt.c_ics in
	let cstrt = { cstrt with c_ics = ics; } in
	let ctx = { ctx with ctx_cstrt = cstrt; } in

	let ctx = set_flags ctx op_size1 inte inte1 inte2 in

	set_integer ctx inte op1

let type_pinst_DEC dnames ctx op =
	let op_size = get_op_size op in
	let inte = get_integer ctx op in

	(* Cannot DEC ESP! *)
	match op with
	| P_Op_Reg (Instructions.Reg32 Instructions.ESP) ->
		failwithf_type_error "@[Stack pointer cannot be DECed.@]"
	| _ ->

	let dname, _ = get_dname dnames in
	let inte', ctx = alloc_inte_tv dname ctx in

	let ie1 = ie_inte inte in
	let ie2 = (ie_inte inte') +.. (ie_int 1) in
	let ic = ie1 =.. ie2 in

	let cstrt = ctx.ctx_cstrt in
	let ics = add_integer_cstrt ic cstrt.c_ics in
	let cstrt = { cstrt with c_ics = ics; } in
	let ctx = { ctx with ctx_cstrt = cstrt; } in

	let ctx = set_flags ctx op_size inte' inte (IConst bi_1) in

	set_integer ctx inte' op

let type_pinst_CMP dbg_new_names ctx op1 op2 =
	let op_size1 = get_op_size op1 in
	let op_size2 = get_op_size op2 in

	(* XXX : should we check operand size ? *)

	let inte1 = get_integer ctx op1 in
	let inte2 = get_integer ctx op2 in

	let inte, ctx = alloc_inte_tv_with_dnames dbg_new_names ctx in

	let inte1 = shift_integer 0 1 inte1 in
	let inte2 = shift_integer 0 1 inte2 in

	let ie1 = ie_inte inte1 in
	let ie2 = (ie_inte inte) +.. (ie_inte inte2) in
	let ic = ie1 =.. ie2 in

	let cstrt = ctx.ctx_cstrt in
	let ics = add_integer_cstrt ic cstrt.c_ics in
	let cstrt = { cstrt with c_ics = ics; } in
	let ctx = { ctx with ctx_cstrt = cstrt; } in

	let ctx = set_flags ctx op_size1 inte inte1 inte2 in

	ctx

let type_pinst_IMUL dnames ctx op1 op2 op3 =
	let inte, ctx = alloc_inte_tv_with_dnames dnames ctx in

	let inte2 = get_integer ctx op2 in
	let inte3 = get_integer ctx op3 in

	let ie1 = ie_inte inte in
	let ie2 = (ie_inte inte2) *.. (ie_inte inte3) in
	let ic = ie1 =.. ie2 in

	let cstrt = ctx.ctx_cstrt in
	let ics = add_integer_cstrt ic cstrt.c_ics in
	let cstrt = { cstrt with c_ics = ics; } in
	let ctx = { ctx with ctx_cstrt = cstrt; } in

	(* XXX : EFLAGS ? *)

	set_integer ctx inte op1

let type_pinst_PUSHF dnames ctx =
	let tv, ctx = get_new_type_variable ctx in
	let lst = [(IVar tv)] in

	Stack.push_things dnames ctx lst

let type_pinst_PUSH dbg_name ctx op =
	let inte = get_integer ctx op in
	let lst = [inte] in

	Stack.push_things dbg_name ctx lst

let type_pinst_POP dbg_name ctx op =
	let lst, ctx = Stack.pop dbg_name ctx 1 in

	if List.length lst <> 1 then
		failwithf_typechecker_bug "We popped one stack element, but two or more stack elements were popped.";

	let inte = List.nth lst 0 in

	(* XXX : we have to shift operand because Stack.pop introduces two new variables *)
	let op = shift_operand 0 2 op in
	let ctx = set_integer ctx inte op in
	ctx

let type_pinst_POPF dbg_name ctx =
	let lst, ctx = Stack.pop dbg_name ctx 1 in

	ctx


let byte_type = Concrete (Pack
{
	pt_prefix =
	{
		p_kinds = [KInt, "__"];
		p_cstrts = [];
(*
		p_cstrts = [ { c_ics = []; c_store = empty_store; c_code_store = empty_code_store; } ];
*)
	};
	pt_tuple = [OpSize8, IVar (TV 0)];
})

let type_pinst_PCAST_TO_TUPLE dnames ctx inte =
try
	let ht, ctx = remove_from_context ctx inte in
	let at = Cast.heap_to_array ht in

	let len = at.at_len in
	match len with
	| IVar _ ->
		failwithf_type_error "Cannot cast the variable-length array."
	| IConst n ->

	let ics = ctx.ctx_cstrt.c_ics in

	if not (is_equal_type ics at.at_type byte_type) then
		failwithf_type_error "Cannot cast non-integer arrays."
	else

(*
	let n = Nativeint.to_int n in
*)
	let n, m = n /%^ bi_size_of_word in

	if m <>^ bi_0 then
		failwithf_type_error "Cannot cast non-aligned arrays."
	else

	let n = Big_int.int_of_big_int n in
	let itvs, ctx = alloc_inte_tvs_with_dnames dnames n ctx in
	let inte = shift_integer 0 n inte in
	let tt = List.map (fun inte -> (OpSize32, inte)) itvs in

	let t = Concrete (Tuple tt) in
	let ht = Array { at_type = t; at_len = IConst bi_1; } in

	let ctx = add_type ctx inte ht in

	ctx
with Cast.Error _ ->
	failwithf_type_error "Cannot cast %a to a tuple."
		Output_types.output_inte (ctx, inte)

let cast_tuple_to_array ctx at inte =
	let len = at.at_len in
	match len with
	| IVar _ ->
		failwithf_type_error "Cannot cast the variable-length array."
	| IConst n ->

(*
	if not (n = 1n) then
*)
	if n <>^ bi_1 then
		failwithf_type_error "Cannot cast the variable-length array."
	else
try
	let tt = Cast.concrete_to_tuple (Cast.typ_to_concrete at.at_type) in
(*
	let len = List.length tt in
*)
	let sz = size_of_tuple_type tt in
	let ht = Array { at_type = byte_type; at_len = IConst (bi sz); } in

	add_type ctx inte ht
with Cast.Error _ ->
	failwithf_type_error "Cannot cast non-tuple arrays."

(*
let rec get_stk_len = function
	| StkNull ->
		0
	| StkCons (_, stk) ->
		4 + (get_stk_len stk)
	| _ ->
		failwithf_type_error "Cannot cast unknown size stacks."
*)

let cast_stack_to_array dnames ctx rstk inte =
	let (new_loc, new_len), ctx = alloc_inte_two_tvs_with_dnames dnames ctx in
	let rstk = shift_rich_stack_type 0 1 rstk in
	let inte = shift_integer 0 1 inte in

	let cstrt = ctx.ctx_cstrt in
	let ics = cstrt.c_ics in

(*
	let ie1 = (ie_inte rstk.rstk_freelen) *.. (ie_inte (IConst bi_size_of_word)) in
*)
	let ie1 = ie_inte rstk.rstk_freelen in
	let ie2 = (ie_inte new_loc) +.. ie1 in
	let ie3 = ie_inte inte in
	let ic = ie2 =.. ie3 in
	let ics = ic :: ics in

	let sz = raw_size_of_stack_type rstk.rstk_stk in

(*
	let ie1 = (ie_inte rstk.rstk_freelen) +.. (ie_int sz) in
	let ie2 = ie1 *.. (ie_inte (IConst bi_size_of_word)) in
*)
	let ie1 = ie_inte rstk.rstk_freelen in
	let ie2 = ie1 +.. (ie_int sz) in
	let ie3 = ie_inte new_len in
	let ic = ie2 =.. ie3 in
	let ics = ic :: ics in

	let cstrt = { cstrt with c_ics = ics; } in
	let ctx = { ctx with ctx_cstrt = cstrt; } in

	let ht = Array { at_type = byte_type; at_len = new_len; } in

	add_type ctx new_loc ht

let type_pinst_PCAST_TO_ARRAY dnames ctx inte =
	let ht, ctx = remove_from_context ctx inte in
	match ht with
	| Array at ->
		cast_tuple_to_array ctx at inte
	| Stack stk ->
		cast_stack_to_array dnames ctx stk inte

let type_pinst_PCAST_TO_STACK dnames ctx inte =
try
	let ht, ctx = remove_from_context ctx inte in
	let at = Cast.heap_to_array ht in

	let len = at.at_len in

	if not (is_equal_type ctx.ctx_cstrt.c_ics at.at_type byte_type) then
		failwithf_type_error "Cannot cast non-byte arrays."
	else

	let new_loc, ctx = alloc_inte_tv_with_dnames dnames ctx in
	let inte = shift_integer 0 1 inte in
	let len = shift_integer 0 1 len in

	let cstrt = ctx.ctx_cstrt in
	let ics = cstrt.c_ics in

	let ie1 = (ie_inte inte) +.. (ie_inte len) in
	let ie2 = ie_inte new_loc in
	let ic = ie1 =.. ie2 in
	let ics = ic :: ics in

	let ht = Stack { rstk_stk = StkNull; rstk_freelen = len; } in

	let cstrt = { cstrt with c_ics = ics; } in
	let ctx = { ctx with ctx_cstrt = cstrt; } in
	let ctx = add_type ctx new_loc ht in

	ctx
with Cast.Error _ ->
	failwithf_type_error "Cannot cast %a to a stack."
		Output_types.output_inte (ctx, inte)

let type_pinst gctx ctx va nva pinst =
	let _, tv_names = List.split ctx.ctx_kinds in
	let tvs = List.map (fun str -> Readable_types.TV(str)) tv_names in
	let pinst = Convert_from_readable.convert_pseudo_inst tvs pinst in

	let c2s = context_to_state in
try
	match pinst with
	| P_MOV (op1, op2) ->
		c2s (type_pinst_MOV ctx op1 op2)
	| P_ADD (dnames, op1, op2) ->
		c2s (type_pinst_ADD dnames ctx op1 op2)
	| P_SUB (dnames, op1, op2) ->
		c2s (type_pinst_SUB dnames ctx op1 op2)
	| P_CMP (dnames, op1, op2) ->
		c2s (type_pinst_CMP dnames ctx op1 op2)
	| P_DEC (dnames, op) ->
		c2s (type_pinst_DEC dnames ctx op)
	| P_IMUL (dnames, op1, op2, op3) ->
		c2s (type_pinst_IMUL dnames ctx op1 op2 op3)
	| P_PUSHF dnames ->
		c2s (type_pinst_PUSHF dnames ctx)
	| P_PUSH (dnames, op) ->
		c2s (type_pinst_PUSH dnames ctx op)
	| P_POP (dnames, op) ->
		c2s (type_pinst_POP dnames ctx op)
	| P_POPF dnames ->
		c2s (type_pinst_POPF dnames ctx)
	| P_PACK (inte, pt, cs) ->
		type_pinst_PPACK ctx inte pt cs
	| P_UNPACK (dnames, inte) ->
		type_pinst_PUNPACK ctx inte dnames
	| P_ROLL (inte, nt) ->
		c2s (type_pinst_PROLL gctx ctx inte nt)
	| P_UNROLL inte ->
		c2s (type_pinst_PUNROLL gctx ctx inte)
	| P_SPLIT (dnames, inte1, inte2) ->
		c2s (type_pinst_PSPLIT dnames ctx inte1 inte2)
	| P_CONCAT (dnames, inte1, inte2) ->
		c2s (type_pinst_PCONCAT dnames ctx inte1 inte2)
	| P_TUPLE_SPLIT (dnames, inte1, inte2) ->
		c2s (type_pinst_PTUPLE_SPLIT dnames ctx inte1 inte2)
	| P_TUPLE_CONCAT (inte1, inte2) ->
		c2s (type_pinst_PTUPLE_CONCAT ctx inte1 inte2)
	| P_Jcc (cc, op) ->
		c2s (type_pinst_Jcc gctx.gctx_code_store ctx nva cc op)
	| P_CALLN op ->
		c2s (type_pinst_CALLN gctx.gctx_code_store ctx nva op)
	| P_JMPN op ->
		c2s (type_pinst_JMPN gctx.gctx_code_store ctx nva op)
	| P_RET ->
		c2s (type_pinst_RET gctx.gctx_code_store ctx)
	| P_FALL ->
		c2s (type_pinst_FALL gctx.gctx_code_store ctx nva)
	| P_CAST_TO_TUPLE (dnames, inte) ->
		c2s (type_pinst_PCAST_TO_TUPLE dnames ctx inte)
	| P_CAST_TO_ARRAY (dnames, inte) ->
		c2s (type_pinst_PCAST_TO_ARRAY dnames ctx inte)
	| P_CAST_TO_STACK (dnames, inte) ->
		c2s (type_pinst_PCAST_TO_STACK dnames ctx inte)
(*
	| _ ->
		failwithf_type_error "Unsupported pseudo instruction"
*)
with Type_error msg ->
		failwithf_type_error "[0x%08nx]: %s" va msg

let is_terminator = function
	| Readable_pseudo_insts.P_CALLN _
	| Readable_pseudo_insts.P_JMPN _
	| Readable_pseudo_insts.P_RET
	| Readable_pseudo_insts.P_FALL ->
		true
	| _ ->
		false

let rec is_closed = function
	| [] ->
		false
	| [(_, _, pinst)] ->
		is_terminator pinst
	| _ :: pinsts ->
		is_closed pinsts

let rec type_pinsts_ctx ntds ctx = function
	| [] ->
		true
	| (va, nva, pinst) :: pinsts ->
	try
		let state = type_pinst ntds ctx va nva pinst in
		type_pinsts_state ntds state pinsts
	with Skip ->
		true

and type_pinsts_state ntds state pinsts =
	let ctxs = state_to_contexts state in
	List.for_all (fun ctx -> type_pinsts_ctx ntds ctx pinsts) ctxs

let type_insts gctx state va iinfos =
        let pinsts = Convert_to_pinsts.iinfos_to_pinsts va iinfos gctx.gctx_annots_map in
	if is_closed pinsts && type_pinsts_state gctx state pinsts then
        	()
	else
		failwithf_type_error "Instruction were not terminated."
