(*
 * Copyright 2005 Toshiyuki Maeda
 *
 * This file is part of TALK.
 *
 * TALK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * TALK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *)

open Types
open Types_op

open Big_int_util

(*
 *
 * Routines for unification
 *
 *)

exception Cannot_unify

module TVMap = Map.Make (Types.TTV)

let is_bound b tv =
	let TV(i) = tv in
	List.exists (fun (m, n) -> m <= i && i < n) b

let shift_bound_vars n b =
	List.map (fun (m1, m2) -> (m1 + n, m2 + n)) b

let shift_subst n s =
	TVMap.fold (fun tv c s ->
		let tv = shift_type_variable 0 n tv in
		let c = shift_constructor 0 n c in
		TVMap.add tv c s
	) s TVMap.empty

let drop_subst n s =
	TVMap.fold (fun tv c s ->
	try
		let tv = shift_type_variable 0 (-n) tv in
		try
			let c = shift_constructor 0 (-n) c in
			TVMap.add tv c s
		with Invalid_shift ->
			raise Cannot_unify
	with Invalid_shift ->
		s
	) s TVMap.empty

let add_subst g_ics tv c s =
	if TVMap.mem tv s then
		let c' = TVMap.find tv s in
		if is_equal_constructor g_ics c c' then
			s
		else
			raise Cannot_unify
	else
		TVMap.add tv c s

let apply_subst_common subst s a =
	TVMap.fold (fun tv c a ->
		subst tv c a
	) s a

let apply_subst_integer s inte = apply_subst_common subst_integer s inte
let apply_subst_register s reg = apply_subst_common subst_register s reg
let apply_subst_tuple_type s tt = apply_subst_common subst_tuple_type s tt
let apply_subst_cstrts s cstrts = apply_subst_common subst_cstrts s cstrts
let apply_subst_store s store = apply_subst_common subst_store s store
let apply_subst_code_store s c_str= apply_subst_common subst_code_store s c_str
let apply_subst_integer_cstrts s ics = apply_subst_common subst_integer_cstrts s ics

let choose_one c_str =
	let r = ref None in
try
	IntMap.iter (fun inte ct ->
		r := Some (inte, ct);
		raise Not_found
	) c_str;
	!r
with _ ->
	!r

let unify_integer g_ics b s inte1 inte2 =

	if is_equal_integer g_ics inte1 inte2 then
		s
	else

	match inte1, inte2 with
	| IVar tv1, _ when is_bound b tv1 ->
		add_subst g_ics tv1 (CInt inte2) s
	| _, IVar tv2 when is_bound b tv2 ->
		add_subst g_ics tv2 (CInt inte1) s
	| _, _ ->
		raise Cannot_unify

let unify_register_type g_ics b s reg1 reg2 =
	let (reg_kind1, inte1) = reg1 in
	let (reg_kind2, inte2) = reg2 in

	if not (is_equal_register_kind reg_kind1 reg_kind2) then
		raise Cannot_unify
	else
		unify_integer g_ics b s inte1 inte2

let unify_flags_type g_ics b s flags1 flags2 =
	if flags1.f_size <> flags2.f_size then
		raise Cannot_unify
	else

	let f_rst1 = apply_subst_integer s flags1.f_rst in
	let f_rst2 = apply_subst_integer s flags2.f_rst in
	let s = unify_integer g_ics b s f_rst1 f_rst2 in

	let f_op11 = apply_subst_integer s flags1.f_op1 in
	let f_op12 = apply_subst_integer s flags2.f_op1 in
	let s = unify_integer g_ics b s f_op11 f_op12 in

	let f_op21 = apply_subst_integer s flags1.f_op2 in
	let f_op22 = apply_subst_integer s flags2.f_op2 in
	let s = unify_integer g_ics b s f_op21 f_op22 in

	s

let unify_flags_type_opt g_ics b s flags_opt1 flags_opt2 =
	match flags_opt1, flags_opt2 with
	| _, None ->
		s
	| None, Some _ ->
		raise Cannot_unify
	| Some flags1, Some flags2 ->
		unify_flags_type g_ics b s flags1 flags2

let unify_registers_type g_ics b s regs1 regs2 =

	if regs1.cs <> regs2.cs then
		raise Cannot_unify
	else

	let s = unify_register_type g_ics b s regs1.eax regs2.eax in

	let ecx1 = apply_subst_register s regs1.ecx in
	let ecx2 = apply_subst_register s regs2.ecx in
	let s = unify_register_type g_ics b s ecx1 ecx2 in

	let edx1 = apply_subst_register s regs1.edx in
	let edx2 = apply_subst_register s regs2.edx in
	let s = unify_register_type g_ics b s edx1 edx2 in

	let ebx1 = apply_subst_register s regs1.ebx in
	let ebx2 = apply_subst_register s regs2.ebx in
	let s = unify_register_type g_ics b s ebx1 ebx2 in

	let esi1 = apply_subst_register s regs1.esi in
	let esi2 = apply_subst_register s regs2.esi in
	let s = unify_register_type g_ics b s esi1 esi2 in

	let edi1 = apply_subst_register s regs1.edi in
	let edi2 = apply_subst_register s regs2.edi in
	let s = unify_register_type g_ics b s edi1 edi2 in

	let ebp1 = apply_subst_register s regs1.ebp in
	let ebp2 = apply_subst_register s regs2.ebp in
	let s = unify_register_type g_ics b s ebp1 ebp2 in

	let esp1 = apply_subst_register s regs1.esp in
	let esp2 = apply_subst_register s regs2.esp in
	let s = unify_register_type g_ics b s esp1 esp2 in

	let s = unify_flags_type_opt g_ics b s regs1.flags_opt regs2.flags_opt in
	s

let get_intersection_slow g_ics lst1 lst2 =
	List.fold_left (fun rst inte1 ->
		List.fold_left (fun rst inte2 ->
			if is_equal_integer g_ics inte1 inte2 then
				(inte1, inte2) :: rst
			else
				rst
		) rst lst2
	) [] lst1

let get_intersection_fast lst1 lst2 =
	List.fold_left (fun rst inte1 ->
		List.fold_left (fun rst inte2 ->
			let b = match inte1, inte2 with
				| IConst i1, IConst i2 ->
					i1 =^ i2
				| IVar tv1, IVar tv2 ->
					is_equal_type_variable tv1 tv2
				| _ ->
					false
			in
			if b then
				(inte1, inte2) :: rst
			else
				rst
		) rst lst2
	) [] lst1

let get_intersection g_ics lst1 lst2 =
	let lst = get_intersection_fast lst1 lst2 in
	match lst with
	| [] ->
		get_intersection_slow g_ics lst1 lst2
	| _ ->
		lst

let inte_is_const = function
	| IConst _, IConst _ ->
		true
	| _ ->
		false

let inte_is_bound_tv b = function
	| IVar tv1, IVar tv2 when is_bound b tv1 && is_bound b tv2 ->
		true
	| _ ->
		false

let get_common_location_opt g_ics b locs1 locs2 =

	let lst1 = IntMap.fold (fun inte _ lst -> inte :: lst ) locs1 [] in
	let lst2 = IntMap.fold (fun inte _ lst -> inte :: lst ) locs2 [] in

	let lst = get_intersection g_ics lst1 lst2 in

	if List.exists inte_is_const lst then
		Some (List.find inte_is_const lst)
	else if List.exists (inte_is_bound_tv b) lst then
		Some (List.find (inte_is_bound_tv b) lst)
	else

	match lst with
	| [] ->
		None
	| p :: _ ->
		Some p

let eliminate_common tvs1 tvs2 =
	let tvs1' = TVSet.diff tvs1 tvs2 in
	let tvs2' = TVSet.diff tvs2 tvs1 in

	tvs1', tvs2'

let get_bound_tv_one b vs =

	let bvs, vs = TVSet.partition (fun tv -> is_bound b tv) vs in

	let n = TVSet.cardinal bvs in

	match n with
	| 0 ->
		None, vs
	| 1 ->
		let tv = TVSet.choose bvs in
		(Some tv), vs
	| _ ->
		raise Cannot_unify


let unify_store_without_unifying_locs g_ics b s str1 str2 =

	let locs1 = str1.str_locs in
	let locs2 = str2.str_locs in

	let (vs1, vs2) = eliminate_common str1.str_vs str2.str_vs in

	if IntMap.equal (is_equal_heap_type g_ics) locs1 locs2 && TVSet.equal vs1 vs2 then
		s
	else

	let tv1_opt, vs1 = get_bound_tv_one b vs1 in
	let tv2_opt, vs2 = get_bound_tv_one b vs2 in

	let str1 = { str_locs = locs1; str_vs = vs1; } in
	let str2 = { str_locs = locs2; str_vs = vs2; } in

	match (tv1_opt, tv2_opt) with
	| None, None ->
		raise Cannot_unify
	| None, Some tv2 ->
		if is_empty_store str2 then
			add_subst g_ics tv2 (CStore str1) s
		else
			raise Cannot_unify
	| Some tv1, None ->
		if is_empty_store str1 then
			add_subst g_ics tv1 (CStore str2) s
		else
			raise Cannot_unify
	| Some tv1, Some tv2 ->
		let s = add_subst g_ics tv2 (CStore str1) s in
		let s = add_subst g_ics tv1 (CStore str2) s in

		s

let rec unify_store g_cstr g_ics b s str1 str2 =

	let locs1 = str1.str_locs in
	let locs2 = str2.str_locs in

	let inte_opt = get_common_location_opt g_ics b locs1 locs2 in

	match inte_opt with
	| None ->
		unify_store_without_unifying_locs g_ics b s str1 str2

	| Some (inte1, inte2) -> (* true = is_equal_integer g_ics inte1 inte2 *)

	let ht1 = IntMap.find inte1 locs1 in
	let ht2 = IntMap.find inte2 locs2 in

	let locs1 = remove_location inte1 locs1 in
	let locs2 = remove_location inte2 locs2 in

	let s = unify_heap_type g_cstr g_ics b s ht1 ht2 in

	let str1 = { str1 with str_locs = locs1; } in
	let str2 = { str2 with str_locs = locs2; } in

	let str1 = apply_subst_store s str1 in
	let str2 = apply_subst_store s str2 in
	let g_ics = apply_subst_integer_cstrts s g_ics in

	unify_store g_cstr g_ics b s str1 str2

and unify_code_store g_cstr g_ics b s c_str1 c_str2 =

	let r_opt = choose_one c_str2 in

	match r_opt with
	| None ->
		s
	| Some (inte, ct2) ->

	let c_str2 = IntMap.remove inte c_str2 in
	let g_cstr' = add_code_store c_str1 g_cstr in

	let ct1 =
	try
		IntMap.find inte g_cstr'
	with Not_found ->
		raise Cannot_unify
	in

	let s = unify_code_type g_cstr' g_ics b s ct2 ct1 in

	let c_str1 = apply_subst_code_store s c_str1 in
	let c_str2 = apply_subst_code_store s c_str2 in
	let g_ics = apply_subst_integer_cstrts s g_ics in

	unify_code_store g_cstr g_ics b s c_str1 c_str2

and unify_code_type g_cstr g_ics b s ct1 ct2 =
	match (ct1, ct2) with
	| NameCode nt1, NameCode nt2 ->
		unify_named_type g_cstr g_ics b s nt1 nt2
	| State st1, State st2 ->
		unify_state_type g_cstr g_ics b s st1 st2
	| _, _ ->
		raise Cannot_unify

and unify_stack_type g_ics b s st1 st2 =
	if is_equal_stack_type g_ics st1 st2 then
		s
	else

	match (st1, st2) with
	| StkNull, StkNull ->
		s
	| StkCons (i1, st1), StkCons(i2, st2) ->
		let s = unify_integer g_ics b s i1 i2 in
		unify_stack_type g_ics b s st1 st2
	| StkVar tv1, _ when is_bound b tv1 ->
		add_subst g_ics tv1 (CStack st2) s
	| _, StkVar tv2 when is_bound b tv2 ->
		add_subst g_ics tv2 (CStack st1) s
	| _, _ ->
		raise Cannot_unify

and unify_rich_stack_type g_ics b s rstk1 rstk2 =
	let s = unify_integer g_ics b s rstk1.rstk_freelen rstk2.rstk_freelen in
	unify_stack_type g_ics b s rstk1.rstk_stk rstk2.rstk_stk

and unify_constructor g_cstr g_ics b s c1 c2 =
	match (c1, c2) with
	| CInt i1, CInt i2 ->
		unify_integer g_ics b s i1 i2
	| CStack st1, CStack st2 ->
		unify_stack_type g_ics b s st1 st2
	| CStore str1, CStore str2 ->
		unify_store g_cstr g_ics b s str1 str2
	| _, _ ->
		raise Cannot_unify

and unify_constructors g_cstr g_ics b s cs1 cs2 =
try
	let cs = List.combine cs1 cs2 in
	List.fold_left (fun s (c1, c2) ->
		unify_constructor g_cstr g_ics b s c1 c2
	) s cs
with Invalid_argument _ ->
	raise Cannot_unify

and unify_named_type g_cstr g_ics b s nt1 nt2 =
	let (n1, cs1) = nt1 in
	let (n2, cs2) = nt2 in

	if (n1 = n2) then
		unify_constructors g_cstr g_ics b s cs1 cs2
	else
		raise Cannot_unify

and unify_pack_type g_cstr g_ics b s pt1 pt2 =

	let kinds1, _ = List.split pt1.pt_prefix.p_kinds in
	let kinds2, _ = List.split pt2.pt_prefix.p_kinds in

	if kinds1 <> kinds2 then
		raise Cannot_unify

	else

	let n = List.length kinds1 in

	let b = shift_bound_vars n b in
	let s = shift_subst n s in

	(* ???: is this really necessary? *)
	let g_cstr = shift_code_store 0 n g_cstr in

	let s = unify_tuple_type g_ics b s pt1.pt_tuple pt2.pt_tuple in

	let cstrts1 = pt1.pt_prefix.p_cstrts in
	let cstrts2 = pt2.pt_prefix.p_cstrts in

	let cstrts1 = apply_subst_cstrts s cstrts1 in
	let cstrts2 = apply_subst_cstrts s cstrts2 in
	let g_ics = apply_subst_integer_cstrts s g_ics in

	let s = unify_cstrts g_cstr g_ics b s cstrts1 cstrts2 in

	let s = drop_subst n s in

	s

and unify_cstrts g_cstr g_ics b s cstrts1 cstrts2 =

	let len1 = List.length cstrts1 in
	let len2 = List.length cstrts2 in

	if len1 <> len2 then
		raise Cannot_unify
	else

	match cstrts1 with
	| [] ->
		s
	| cstrt1 :: cstrts1 ->

	let s, cstrts2 = unify_cstrts_one g_cstr g_ics b s cstrt1 cstrts2 in

	let cstrts1 = apply_subst_cstrts s cstrts1 in
	let cstrts2 = apply_subst_cstrts s cstrts2 in
	let g_ics = apply_subst_integer_cstrts s g_ics in

	unify_cstrts g_cstr g_ics b s cstrts1 cstrts2

and unify_and_check_cstrts g_cstr g_ics b s cstrts1 cstrts2 =

	match cstrts1 with
	| [] ->
		s
	| cstrt1 :: cstrts1 ->
	let s = unify_and_check_cstrts_one g_cstr g_ics b s cstrt1 cstrts2 in

	let cstrts1 = apply_subst_cstrts s cstrts1 in
	let cstrts2 = apply_subst_cstrts s cstrts2 in
	let g_ics = apply_subst_integer_cstrts s g_ics in

	let s = unify_and_check_cstrts g_cstr g_ics b s cstrts1 cstrts2 in
	s

and unify_cstrts_one g_cstr g_ics b s cstrt cstrts =
	unify_cstrts_one_aux g_cstr g_ics b s cstrt cstrts []

and unify_cstrts_one_aux g_cstr g_ics b s cstrt1 cstrts r_cstrts =
	match cstrts with
	| [] ->
		raise Cannot_unify
	| cstrt2 :: cstrts ->
	try
		let s = unify_cstrt g_cstr g_ics b s cstrt1 cstrt2 in
		(s, r_cstrts)
	with Cannot_unify ->
		unify_cstrts_one_aux g_cstr g_ics b s cstrt1 cstrts (cstrt2 :: r_cstrts)

and unify_and_check_cstrts_one g_cstr g_ics b s cstrt1 cstrts =

	match cstrts with
	| [] ->
		raise Cannot_unify
	| cstrt2 :: cstrts ->
	try
		let s = unify_and_check_cstrt g_cstr g_ics b s cstrt1 cstrt2 in
		s
	with Cannot_unify ->
		unify_and_check_cstrts_one g_cstr g_ics b s cstrt1 cstrts

and unify_cstrt g_cstr g_ics b s cstrt1 cstrt2 =

	let s = unify_store g_cstr g_ics b s cstrt1.c_store cstrt2.c_store in

	let c_str1 = apply_subst_code_store s cstrt1.c_code_store in
	let c_str2 = apply_subst_code_store s cstrt2.c_code_store in
	let g_ics = apply_subst_integer_cstrts s g_ics in

	let s = unify_code_store g_cstr g_ics b s c_str1 c_str2 in

	(* check integer constraints *)
	let ics1 = apply_subst_integer_cstrts s cstrt1.c_ics in
	let ics2 = apply_subst_integer_cstrts s cstrt2.c_ics in

	if not (is_equal_integer_cstrts g_ics ics1 ics2) then
		raise Cannot_unify
	else
		s

and unify_and_check_cstrt g_cstr g_ics' b s cstrt1 cstrt2 =

	let g_ics = add_integer_cstrts cstrt1.c_ics g_ics' in

	let str1 = cstrt1.c_store in
	let str2 = cstrt2.c_store in

	let s =
try
	Util.dprintf "@[unify_store@]@.";
	let s = unify_store g_cstr g_ics b s str1 str2 in
	Util.dprintf "@[unify_store done.@]@.";
	s
with _ ->
(
	Util.dprintf "@[unify_store failed.@]@.";
	Util.dprintf "@[ics@ =@ %a@]@." Output_types_raw.output_inte_cstrts g_ics;
	Util.dprintf "@[str1@ =@ %a@]@." Output_types_raw.output_store str1;
	Util.dprintf "@[str2@ =@ %a@]@." Output_types_raw.output_store str2;
raise Cannot_unify) in

	let c_str1 = apply_subst_code_store s cstrt1.c_code_store in
	let c_str2 = apply_subst_code_store s cstrt2.c_code_store in
	let g_ics = apply_subst_integer_cstrts s g_ics in

	let s =
try
	Util.dprintf "@[unify_code_store@]@.";
	let s = unify_code_store g_cstr g_ics b s c_str1 c_str2 in
	Util.dprintf "@[unify_code_store done.@]@.";
	s
with _ ->
(
	Util.dprintf "@[unify_code_store failed.@]@.";
	Util.dprintf "@[ics@ =@ %a@]@." Output_types_raw.output_inte_cstrts g_ics;
	Util.dprintf "@[c_str1@ =@ %a@]@." Output_types_raw.output_code_store c_str1;
	Util.dprintf "@[c_str2@ =@ %a@]@." Output_types_raw.output_code_store c_str2;
raise Cannot_unify) in

	(* check integer constraints *)
	let g_ics = apply_subst_integer_cstrts s g_ics in
	let ics2 = apply_subst_integer_cstrts s cstrt2.c_ics in

	if not (Constraint_solver.implies g_ics ics2) then
(
		let g_ics' = apply_subst_integer_cstrts s g_ics' in
		let ics1 = apply_subst_integer_cstrts s cstrt1.c_ics in
		Util.dprintf "@[%a@]" Output_types_raw.output_inte_cstrts g_ics';
		Util.dprintf "@[&@]@.";
		Util.dprintf "@[%a@]" Output_types_raw.output_inte_cstrts ics1;
		Util.dprintf "@[==>@]@.";
		Util.dprintf "@[%a@]" Output_types_raw.output_inte_cstrts ics2;
		raise Cannot_unify
)
	else
		s


and unify_tuple_type g_ics b s tt1 tt2 =

	match tt1, tt2 with
	| [], [] ->
		s
	| te1 :: tt1, te2 :: tt2 ->
		let s = unify_tuple_elem_type g_ics b s te1 te2 in
		let tt1 = apply_subst_tuple_type s tt1 in
		let tt2 = apply_subst_tuple_type s tt2 in
		unify_tuple_type g_ics b s tt1 tt2
	| _, _ ->
		raise Cannot_unify

and unify_tuple_elem_type g_ics b s (sz1, inte1) (sz2, inte2) =
	if sz1 <> sz2 then
		raise Cannot_unify
	else
	unify_integer g_ics b s inte1 inte2

and unify_concrete_type g_cstr g_ics b s ct1 ct2 =
	match (ct1, ct2) with
	| Pack pt1, Pack pt2 ->
		unify_pack_type g_cstr g_ics b s pt1 pt2
	| Tuple tt1, Tuple tt2 ->
		unify_tuple_type g_ics b s tt1 tt2
	| _, _ ->
		raise Cannot_unify

and unify_type g_cstr g_ics b s t1 t2 =
	match (t1, t2) with
	| Name nt1, Name nt2 ->
		unify_named_type g_cstr g_ics b s nt1 nt2
	| Concrete ct1, Concrete ct2 ->
		unify_concrete_type g_cstr g_ics b s ct1 ct2
	| _, _ ->
		raise Cannot_unify

and unify_array_type g_cstrt g_ics b s at1 at2 =
	let { at_type = t1; at_len = len1; } = at1 in
	let { at_type = t2; at_len = len2; } = at2 in

	let s = unify_type g_cstrt g_ics b s t1 t2 in
	let s = unify_type g_cstrt g_ics b s t1 t2 in

	let len1 = apply_subst_integer s len1 in
	let len2 = apply_subst_integer s len2 in

	unify_integer g_ics b s len1 len2

and unify_heap_type g_cstrt g_ics b s ht1 ht2 =
	match ht1, ht2 with
	| Array at1, Array at2 ->
		unify_array_type g_cstrt g_ics b s at1 at2
	| Stack rstk1, Stack rstk2 ->
		unify_rich_stack_type g_ics b s rstk1 rstk2
	| _ ->
		raise Cannot_unify

(*
 * let unify_state type b s state1 state2 =
 *
 *    subst where state1 == [subst]state2
 *)
and unify_state_type g_cstr g_ics b s state1 state2 =

	let regs1 = state1.st_registers in
	let prefix1 = state1.st_prefix in
	let cstrts1 = prefix1.p_cstrts in

	let regs2 = state2.st_registers in
	let prefix2 = state2.st_prefix in
	let cstrts2 = prefix2.p_cstrts in

	let n1 = List.length prefix1.p_kinds in
	let n2 = List.length prefix2.p_kinds in

	let n = n1 + n2 in

	let regs1 = shift_registers_type 0 n2 regs1 in
	let cstrts1 = shift_cstrts 0 n2 cstrts1 in

	let regs2 = shift_registers_type n n1 regs2 in
	let cstrts2 = shift_cstrts n n1 cstrts2 in

	let b = shift_bound_vars n b in
	let s = shift_subst n s in

	let g_cstr = shift_code_store 0 (n1 + n2) g_cstr in
	let g_ics = shift_integer_cstrts 0 (n1 + n2) g_ics in

	let b' = (0, n2) :: b in

	let s = unify_registers_type g_ics b' s regs1 regs2 in

	let cstrts1 = apply_subst_cstrts s cstrts1 in
	let cstrts2 = apply_subst_cstrts s cstrts2 in
	let g_ics = apply_subst_integer_cstrts s g_ics in

	let s = unify_and_check_cstrts g_cstr g_ics b' s cstrts1 cstrts2 in

	let s = drop_subst (n1 + n2) s in

	s

let is_subtype_state_type g_cstr ctx state2 =
	let state1 = Types_op.context_to_state ctx in
	let ics = [] in
try
	let _ = unify_state_type g_cstr ics [] TVMap.empty state1 state2 in

	true
with Cannot_unify ->
	false
